#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.net.rfb.d3des import generate_response, deskey, desfunc, decrypt_passwd

class TestD3DES(unittest.TestCase):

    def test_des(self):
        key = bytearray.fromhex('0123456789abcdef')
        plain = bytearray.fromhex('0123456789abcdef')
        cipher = bytearray.fromhex('6e09a37726dd560c')
        ek = deskey(key, False)
        dk = deskey(key, True)
        assert desfunc(plain, ek) == cipher
        assert desfunc(desfunc(plain, ek), dk) == plain
        assert desfunc(desfunc(plain, dk), ek) == plain

    def test_generate_response(self):
        challange = b"helloworld0123456789"[:16]
        for passwd in (b"", b"0"*32):
            r = generate_response(passwd, challange)
            assert r
        assert decrypt_passwd(b"helloworld"[:8])

def main():
    unittest.main()

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import typedict, AdHocStruct
from xpra.util.thread import start_thread
from xpra.server.mixins.stub_server_mixin import StubServerMixin


class EncodingMixinTest(unittest.TestCase):

    def test_mixin_methods(self):
        opts = AdHocStruct()
        x = StubServerMixin()
        x.init(opts)
        x.init_state()
        x.init_sockets([])
        x.setup()
        t = start_thread(x.threaded_setup, "threaded setup")
        t.join()
        x.init_packet_handlers()
        x.get_server_source(None)

        assert isinstance(x.get_caps(None), dict)
        assert isinstance(x.get_server_features(None), dict)
        assert isinstance(x.get_info(None), dict)
        assert isinstance(x.get_ui_info(None, None), dict)

        caps = typedict()
        x.parse_hello(None, caps, True)
        x.add_new_client(None, caps, True, 1)
        x.send_initial_data(None, caps, True, 1)

        x.last_client_exited()
        x.set_session_driver(None)

        x.cleanup_protocol(None)
        x.cleanup()


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import hashlib
from typing import Any

from xpra.util.objects import typedict
from xpra.util.str_fn import csv, Ellipsizer
from xpra.util.env import envbool
from xpra.os_util import gi_import
from xpra.gtk.keymap import get_gtk_keymap, get_default_keymap
from xpra.gtk.error import xsync, xlog
from xpra.keyboard.mask import (
    DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_MEANINGS, DEFAULT_MODIFIER_NUISANCE_KEYNAMES, mask_to_names,
)
from xpra.server.keyboard_config_base import KeyboardConfigBase
from xpra.x11.gtk.keys import grok_modifier_map
from xpra.x11.xkbhelper import (
    do_set_keymap, set_all_keycodes, set_keycode_translation,
    get_modifiers_from_meanings, get_modifiers_from_keycodes,
    clear_modifiers, set_modifiers, map_missing_modifiers,
    clean_keyboard_state, get_keycode_mappings,
    DEBUG_KEYSYMS,
)
from xpra.x11.bindings.keyboard import X11KeyboardBindings
from xpra.log import Logger

log = Logger("keyboard")
verboselog = Logger("keyboard", "verbose")

Gdk = gi_import("Gdk")

X11Keyboard = X11KeyboardBindings()

MAP_MISSING_MODIFIERS: bool = envbool("XPRA_MAP_MISSING_MODIFIERS", True)
SHIFT_LOCK: bool = envbool("XPRA_SHIFT_LOCK", False)
VERIFY_MODIFIERS: bool = envbool("XPRA_VERIFY_MODIFIERS", True)

ALL_X11_MODIFIERS: dict[str, int] = {
    "shift": 0,
    "lock": 1,
    "control": 2,
    "mod1": 3,
    "mod2": 4,
    "mod3": 5,
    "mod4": 6,
    "mod5": 7,
}


def kmlog(keyname: str, msg: str, *args) -> None:
    """ logs at info level is we are debugging this keyname """
    if keyname in DEBUG_KEYSYMS:
        lfn = log.info
    else:
        lfn = log
    lfn(msg, *args)


class KeyboardConfig(KeyboardConfigBase):
    def __init__(self):
        super().__init__()
        self.raw: bool = False
        self.query_struct = None
        self.modifier_map: dict[str, int] = {}
        self.mod_meanings = {}
        self.mod_managed: list[str] = []
        self.mod_pointermissing: list[str] = []
        self.mod_nuisance: set[str] = set(DEFAULT_MODIFIER_NUISANCE)
        self.keycodes = ()
        self.x11_keycodes = []
        self.layout = ""
        self.variant = ""
        self.options = ""
        self.layout_groups = False

        # this is shared between clients!
        self.keys_pressed = {}
        # these are derived by calling set_keymap:
        self.keynames_for_mod: dict[str, list[str]] = {}
        self.keycode_translation = {}
        self.keycodes_for_modifier_keynames: dict[str, list[int]] = {}
        self.modifier_client_keycodes: dict[str, list] = {}
        self.compute_modifier_map()
        self.keycode_mappings = {}

    def __repr__(self):
        return "KeyboardConfig(%s / %s / %s)" % (self.layout, self.variant, self.options)

    def get_info(self) -> dict[str, Any]:
        info = super().get_info()
        # keycodes:
        if self.keycode_translation:
            ksinfo = info.setdefault("keysym", {})
            kssinf = info.setdefault("keysyms", {})
            kcinfo = info.setdefault("keycode", {})
            for kc, keycode in self.keycode_translation.items():
                if isinstance(kc, tuple):
                    a, b = kc
                    if isinstance(a, int):
                        client_keycode, keysym = a, b
                        ksinfo.setdefault(keysym, {})[client_keycode] = keycode
                        kcinfo.setdefault(client_keycode, {})[keysym] = keycode
                    elif isinstance(b, int):
                        keysym, index = a, b
                        kssinf.setdefault(keycode, []).append((index, keysym))
                else:
                    kcinfo[kc] = keycode
        if self.keycodes:
            i = 0
            kminfo = info.setdefault("keymap", {})
            for keyval, name, keycode, group, level in self.keycodes:
                kminfo[i] = (keyval, name, keycode, group, level)
                i += 1
        # modifiers:
        modinfo: dict[str, dict[str, Any]] = {}
        if self.modifier_client_keycodes:
            for mod, keys in self.modifier_client_keycodes.items():
                modinfo.setdefault(mod, {})["client_keys"] = keys
        if self.keynames_for_mod:
            for mod, keys in self.keynames_for_mod.items():
                modinfo.setdefault(mod, {})["keys"] = tuple(keys)
        if self.keycodes_for_modifier_keynames:
            for mod, keys in self.keycodes_for_modifier_keynames.items():
                modinfo.setdefault(mod, {})["keycodes"] = tuple(keys)
        if self.mod_meanings:
            for mod, mod_name in self.mod_meanings.items():
                modinfo[mod] = mod_name
        info["x11_keycode"] = self.x11_keycodes
        for x in ("layout", "variant", "options", "mod_managed", "mod_pointermissing", "raw", "layout_groups"):
            v = getattr(self, x)
            if v:
                info[x] = v
        modsinfo = {
            "nuisance": tuple(self.mod_nuisance or []),
        }
        info.update({
            "modifier": modinfo,
            "modifiers": modsinfo,
            "keys-pressed": self.keys_pressed,
        })
        # this would need to always run in the UI thread:
        # info["state"] = {
        #    "modifiers" : self.get_current_mask(),
        #    }
        log("keyboard info: %s", Ellipsizer(info))
        verboselog("full keyboard info: %s", info)
        return info

    def parse_options(self, props: typedict) -> int:
        """ used by both process_hello and process_keymap
            to set the keyboard attributes """
        super().parse_options(props)
        modded = {}
        # clients version 4.4 and later use a 'keymap' substructure:
        keymap_dict = typedict(props.dictget("keymap") or {})

        def parse_option(name, parse_fn, old_parse_fn, *parse_args):
            cv = getattr(self, name)
            if keymap_dict:
                nv = parse_fn(name, *parse_args)
            else:
                nv = old_parse_fn("xkbmap_%s" % name, *parse_args)
            if cv != nv:
                setattr(self, name, nv)
                modded[name] = nv

        # lists:
        parse_option("keycodes", keymap_dict.tupleget, props.tupleget)
        # dicts:
        for x in ("mod_meanings", "x11_keycodes", "query_struct"):
            parse_option(x, keymap_dict.dictget, props.dictget, {})
        # lists of strings:
        for x in ("mod_managed", "mod_pointermissing"):
            parse_option(x, keymap_dict.strtupleget, props.strtupleget)
        parse_option("raw", keymap_dict.boolget, props.boolget)
        parse_option("layout_groups", keymap_dict.boolget, props.boolget, True)
        log("parse_options(..) modified %s", Ellipsizer(modded))
        return len(modded)

    def parse_layout(self, props: typedict) -> None:
        """ used by both process_hello and process_keymap """
        keymap_dict = typedict(props.dictget("keymap") or {})
        self.layout = keymap_dict.strget("layout", "us")
        self.variant = keymap_dict.strget("variant")
        self.options = keymap_dict.strget("options")

    def get_hash(self) -> str:
        """
            This hash will be different whenever the keyboard configuration changes.
        """
        m = hashlib.sha256()

        def hashadd(v) -> None:
            m.update(("/%s" % str(v)).encode("utf8"))

        hashadd(super().get_hash())
        for x in (self.raw, self.mod_meanings, self.mod_pointermissing, self.keycodes, self.x11_keycodes):
            hashadd(x)
        if self.query_struct:
            # flatten the dict in a predicatable order:
            for k in sorted(self.query_struct.keys()):
                hashadd(self.query_struct.get(k))
        return "%s/%s/%s/%s" % (self.layout, self.variant, self.options, m.hexdigest())

    def compute_modifiers(self) -> None:
        log("compute_modifiers() raw=%s", self.raw)
        if self.raw:
            with xsync:
                mod_mappings = X11Keyboard.get_modifier_mappings()
            self.mod_meanings = {}
            self.keycodes_for_modifier_keynames = {}
            for mod, mod_defs in mod_mappings.items():
                for mod_def in mod_defs:
                    for v in mod_def:
                        if isinstance(v, int):
                            l = self.keycodes_for_modifier_keynames.setdefault(mod, [])
                        else:
                            self.mod_meanings[v] = mod
                            l = self.keynames_for_mod.setdefault(mod, [])
                        if v not in l:
                            l.append(v)
        else:
            verboselog("compute_modifiers() mod_meanings=%s, keycodes=%s",
                       Ellipsizer(self.mod_meanings), Ellipsizer(self.keycodes))
            verboselog("compute_modifiers() mod_meanings=%s", self.mod_meanings)
            verboselog("compute_modifiers() keycodes=%s", self.keycodes)
            if self.mod_meanings:
                # Unix-like OS provides modifier meanings:
                self.keynames_for_mod = get_modifiers_from_meanings(self.mod_meanings)
            elif self.keycodes:
                # non-Unix-like OS provides just keycodes for now:
                self.keynames_for_mod = get_modifiers_from_keycodes(self.keycodes, MAP_MISSING_MODIFIERS)
                if MAP_MISSING_MODIFIERS:
                    map_missing_modifiers(self.keynames_for_mod)
            else:
                log.info("client did not supply any modifier definitions, using defaults")
                self.keynames_for_mod = get_modifiers_from_meanings(DEFAULT_MODIFIER_MEANINGS)
                if MAP_MISSING_MODIFIERS:
                    map_missing_modifiers(self.keynames_for_mod)
        log("compute_modifiers() keynames_for_mod=%s", self.keynames_for_mod)
        log("compute_modifiers() keycodes_for_modifier_keynames=%s", self.keycodes_for_modifier_keynames)
        log("compute_modifiers() mod_meanings=%s", self.mod_meanings)

    def compute_modifier_keynames(self) -> None:
        self.keycodes_for_modifier_keynames = {}
        self.mod_nuisance = set(DEFAULT_MODIFIER_NUISANCE)
        keymap = get_default_keymap()
        if self.keynames_for_mod:
            for modifier, keynames in self.keynames_for_mod.items():
                for keyname in keynames:
                    if keyname in DEFAULT_MODIFIER_NUISANCE_KEYNAMES:
                        self.mod_nuisance.add(modifier)
                    keyval = Gdk.keyval_from_name(keyname)
                    if keyval == 0:
                        log.error("Error: no keyval found for keyname '%s' (modifier '%s')", keyname, modifier)
                        continue
                    entries = keymap.get_entries_for_keyval(keyval)
                    if entries:
                        keycodes = []
                        if entries[0] is True:
                            keycodes = [entry.keycode for entry in entries[1]]
                        for keycode in keycodes:
                            l = self.keycodes_for_modifier_keynames.setdefault(keyname, [])
                            if keycode not in l:
                                l.append(keycode)
        log("compute_modifier_keynames: keycodes_for_modifier_keynames=%s",
            self.keycodes_for_modifier_keynames)

    def compute_client_modifier_keycodes(self) -> None:
        """ The keycodes for all modifiers (those are *client* keycodes!) """
        with log.trap_error("Error computing client modifier keycodes"):
            server_mappings = X11Keyboard.get_modifier_mappings()
            log("compute_client_modifier_keycodes() server mappings=%s", server_mappings)
            # update the mappings to use the keycodes the client knows about:
            reverse_trans = {}
            for k, v in self.keycode_translation.items():
                reverse_trans[v] = k
            self.modifier_client_keycodes = {}
            self.mod_nuisance = set(DEFAULT_MODIFIER_NUISANCE)
            for modifier, keys in server_mappings.items():
                # ie: modifier=mod3, keys=[(115, 'Super_L'), (116, 'Super_R'), (127, 'Super_L')]
                client_keydefs = []
                for keycode, keysym in keys:
                    # ie: keycode=115, keysym=Super_L
                    client_def = reverse_trans.get(keycode, (0, keysym))
                    # ie: client_def = (99, Super_L)
                    # ie: client_def = Super_L
                    # ie: client_def = (0, Super_L)
                    if isinstance(client_def, (list, tuple)):
                        # ie:
                        # keycode, keysym:
                        #  client_def = (99, Super_L)
                        # or keysym, level:
                        #  client_def = (Super_L, 1)
                        client_keydefs.append(client_def)
                    elif client_def == keysym:
                        # ie: client_def = Super_L
                        client_keydefs.append((keycode, keysym))
                    # record nuisacnde modifiers:
                    if keysym in DEFAULT_MODIFIER_NUISANCE_KEYNAMES:
                        self.mod_nuisance.add(modifier)
                self.modifier_client_keycodes[modifier] = client_keydefs
            log("compute_client_modifier_keycodes() mappings=%s", self.modifier_client_keycodes)
            log("compute_client_modifier_keycodes() mod nuisance=%s", self.mod_nuisance)

    def compute_modifier_map(self) -> None:
        with xlog:
            self.modifier_map = grok_modifier_map(self.mod_meanings)
        log("modifier_map(%s)=%s", self.mod_meanings, self.modifier_map)

    def is_modifier(self, keycode: int) -> bool:
        for mod, keys in self.keycodes_for_modifier_keynames.items():
            if keycode in keys:
                log("is_modifier(%s) found modifier: %s", keycode, mod)
                return True
        log("is_modifier(%s) not found", keycode)
        return False

    def set_layout(self, layout: str, variant: str, options: str) -> bool:
        log("set_layout(%s, %s, %s)", layout, variant, options)
        if layout == "en":
            log.warn("Warning: invalid keyboard layout name '%s', using 'us' instead", layout)
            layout = "us"
        if layout != self.layout or variant != self.variant or options != self.options:
            self.layout = layout
            self.variant = variant
            self.options = options
            return True
        return False

    def set_keymap(self, translate_only=False) -> None:
        if not self.enabled:
            return
        log("set_keymap(%s) layout=%r, variant=%r, options=%r, query-struct=%r",
            translate_only, self.layout, self.variant, self.options, self.query_struct)
        if translate_only:
            self.keycode_translation = set_keycode_translation(self.x11_keycodes, self.keycodes)
            self.add_gtk_keynames()
            self.add_loose_matches()
            self.compute_modifiers()
            self.compute_modifier_keynames()
            self.compute_client_modifier_keycodes()
            self.update_keycode_mappings()
            return

        with xlog:
            clean_keyboard_state()
            do_set_keymap(self.layout, self.variant, self.options, self.query_struct)
        log("set_keymap: query_struct=%r", self.query_struct)
        with xlog:
            # first clear all existing modifiers:
            clean_keyboard_state()

            if not self.raw:
                clear_modifiers()
                clean_keyboard_state()

                # now set all the keycodes:
                # first compute the modifier maps as this may have an influence
                # on the keycode mappings (at least for the from_keycodes case):
                self.compute_modifiers()
                # key translation:
                if self.x11_keycodes and self.query_struct:
                    # native full mapping of all keycodes:
                    self.keycode_translation = set_all_keycodes(self.x11_keycodes, self.keycodes,
                                                                False, self.keynames_for_mod)
                elif self.keycodes:
                    # if the client does not provide a full native keymap with all the keycodes,
                    # try to preserve the initial server keycodes and translate the client keycodes instead:
                    # (used by non X11 clients like osx,win32 or HTML5)
                    self.keycode_translation.update(set_keycode_translation(self.x11_keycodes, self.keycodes))
                else:
                    self.keycode_translation = {}
                self.add_gtk_keynames()

                # now set the new modifier mappings:
                clean_keyboard_state()
                log("going to set modifiers, mod_meanings=%s, len(keycodes)=%s, keynames_for_mod=%s",
                    self.mod_meanings, len(self.keycodes or []), self.keynames_for_mod)
                if self.keynames_for_mod:
                    set_modifiers(self.keynames_for_mod)
                log("keynames_for_mod=%s", self.keynames_for_mod)
                self.compute_modifier_keynames()
            else:
                self.keycode_translation = {}
                log("keyboard raw mode, keycode translation left empty")
                self.compute_modifiers()

            self.compute_client_modifier_keycodes()
            log("keyname_for_mod=%s", self.keynames_for_mod)
            clean_keyboard_state()
            self.update_keycode_mappings()
            self.add_loose_matches()

    def add_gtk_keynames(self) -> None:
        # add the keynames we find via gtk
        # since we may rely on finding those keynames from the client
        # (used with non-native keymaps)
        keymap = get_gtk_keymap()
        log("add_gtk_keynames() gtk keymap=%s", Ellipsizer(keymap))
        for _, keyname, keycode, _, _ in keymap:
            if keyname not in self.keycode_translation:
                self.keycode_translation[keyname] = keycode
                if keyname in DEBUG_KEYSYMS:
                    log.info("add_gtk_keynames: %s=%s", keyname, keycode)

    def add_loose_matches(self) -> None:
        # add lowercase versions of all keynames
        for key, keycode in tuple(self.keycode_translation.items()):
            if isinstance(key, str) and not key.islower():
                self.keycode_translation[key.lower()] = keycode

    def set_default_keymap(self) -> None:
        """
        assign a default keymap based on the current X11 server keymap
        sets up the translation tables,
        so we can look up keys without setting a client keymap.
        """
        if not self.enabled:
            return
        with xsync:
            clean_keyboard_state()
            # keycodes:
            keycode_to_keynames = get_keycode_mappings()
            self.keycode_translation = {}
            # prefer keycodes that don't use the lowest level+mode:
            default_for_keyname: dict[str, tuple[str | int, int]] = {}
            for keycode, keynames in keycode_to_keynames.items():
                for i, keyname in enumerate(keynames):
                    self.keycode_translation[(keyname, i)] = keycode
                    if keyname in DEBUG_KEYSYMS:
                        log.info("set_default_keymap: %s=%s", (keyname, i), keycode)
                    kd = default_for_keyname.get(keyname)
                    if kd is None or kd[1] > i:
                        default_for_keyname[keyname] = (keycode, i)
            for keyname, kd in default_for_keyname.items():
                keycode = kd[0]
                self.keycode_translation[keyname] = keycode
                if keyname in DEBUG_KEYSYMS:
                    log.info("set_default_keymap: %s=%s", keyname, keycode)
            self.add_gtk_keynames()
            log("set_default_keymap: keycode_translation=%s", self.keycode_translation)
            # modifiers:
            self.keynames_for_mod = {}
            # ie: {'control': [(37, 'Control_L'), (105, 'Control_R')], ...}
            mod_mappings = X11Keyboard.get_modifier_mappings()
            log("set_default_keymap: using modifier mappings=%s", mod_mappings)
            for modifier, mappings in mod_mappings.items():
                keynames = []
                for m in mappings:  # ie: (37, 'Control_L'), (105, 'Control_R')
                    if len(m) == 2:
                        keynames.append(m[1])  # ie: 'Control_L'
                self.keynames_for_mod[modifier] = set(keynames)
            self.compute_modifier_keynames()
            self.compute_client_modifier_keycodes()
            log("set_default_keymap: keynames_for_mod=%s", self.keynames_for_mod)
            log("set_default_keymap: keycodes_for_modifier_keynames=%s", self.keycodes_for_modifier_keynames)
            log("set_default_keymap: modifier_map=%s", self.modifier_map)
            self.update_keycode_mappings()

    def update_keycode_mappings(self) -> None:
        self.keycode_mappings = get_keycode_mappings()

    def do_get_keycode(self, client_keycode: int, keyname: str, pressed: bool, modifiers, keyval: int,
                       keystr: str, group: int) -> tuple[int, int]:
        if not self.enabled:
            kmlog(keyname, "ignoring keycode since keyboard is turned off")
            return -1, group
        if keyname == "0xffffff":
            return -1, group
        if self.raw:
            return client_keycode, group
        if self.x11_keycodes and client_keycode > 0:
            keycode = self.keycode_translation.get((client_keycode, keyname), 0) or client_keycode
            kmlog(keyname, "do_get_keycode (%i, %s)=%s (native keymap)", client_keycode, keyname, keycode)
            return keycode, group
        return self.find_matching_keycode(client_keycode, keyname, pressed, modifiers, keyval, keystr, group)

    def find_matching_keycode(self, client_keycode: int, keyname: str,
                              pressed: bool, modifiers, keyval: int, keystr: str, group: int) -> tuple[int, int]:
        """
        from man xmodmap:
        The list of keysyms is assigned to the indicated keycode (which may be specified in decimal,
        hex or octal and can be determined by running the xev program).
        Up to eight keysyms may be attached to a key, however the last four are not used in any major
        X server implementation.
        The first keysym is used when no modifier key is pressed in conjunction with this key,
        the second with Shift, the third when the Mode_switch key is used with this key and
        the fourth when both the Mode_switch and Shift keys are used.
        """
        keycode = -1
        rgroup = group

        def klog(msg: str, *args) -> None:
            kmlog(keyname, "do_get_keycode%s" + msg,
                  (client_keycode, keyname, pressed, modifiers, keyval, keystr, group), *args)

        def kml(msg: str, *args) -> None:
            kmlog(keyname, msg, *args)

        # non-native: try harder to find matching keysym
        # first, try to honour shift state:
        lock = ("lock" in modifiers) and (SHIFT_LOCK or (bool(keystr) and keystr.isalpha()))
        shift = ("shift" in modifiers) ^ lock
        mode = 0
        numlock = 0
        numlock_modifier = None
        for mod, keynames in self.keynames_for_mod.items():
            if "Num_Lock" in keynames:
                numlock_modifier = mod
                break
        for mod in modifiers:
            names = self.keynames_for_mod.get(mod, [])
            if "Num_Lock" in names:
                numlock = 1
            for name in names:
                if name in ("ISO_Level3_Shift", "Mode_switch"):
                    mode = 1
                    break
        levels = []
        # try to preserve the mode (harder to toggle):
        for m in (int(bool(mode)), int(not mode)):
            # try to preserve shift state:
            for s in (int(bool(shift)), int(not shift)):
                # group is comparatively easier to toggle (one function call):
                for g in (int(bool(group)), int(not group)):
                    level = int(g) * 4 + int(m) * 2 + int(s) * 1
                    levels.append(level)
        kml("will try levels: %s", levels)
        for level in levels:
            keycode = self.keycode_translation.get((keyname, level), -1)
            if keycode < 0:
                continue
            keysyms = self.keycode_mappings.get(keycode)
            klog("=%i (level=%i, shift=%s, mode=%i, keysyms=%s)", keycode, level, shift, mode, keysyms)
            level0 = levels[0]
            uq_keysyms = set(keysyms)
            if len(uq_keysyms) <= 1 or (len(keysyms) > level0 and keysyms[level0] == ""):
                # if the keysym we would match for this keycode is 'NoSymbol',
                # then we can probably ignore it ('NoSymbol' shows up as "")
                # same if there's only one actual keysym for this keycode
                kml("not toggling any modifiers state for keysyms=%s", keysyms)
                break

            def toggle_modifier(mod: str) -> None:
                keynames = self.keynames_for_mod.get(mod, set())
                if keyname in keynames:
                    kml("not toggling '%s' since '%s' should deal with it", mod, keyname)
                    # the keycode we're returning is for this modifier,
                    # assume that this will end up doing what is needed
                    return
                if mod in modifiers:
                    kml(f"removing {mod!r} from modifiers")
                    modifiers.remove(mod)
                else:
                    kml(f"adding {mod} to modifiers")
                    modifiers.append(mod)

            # keypad overrules shift state (see #2702):
            if keyname.startswith("KP_"):
                if numlock_modifier and not numlock:
                    toggle_modifier(numlock_modifier)
            elif (level & 1) ^ shift:
                # shift state does not match
                toggle_modifier("shift")
            if int(bool(level & 2)) ^ mode:
                # try to set / unset mode:
                for mod, keynames in self.keynames_for_mod.items():
                    if "ISO_Level3_Shift" in keynames or "Mode_switch" in keynames:
                        # found mode switch modified
                        toggle_modifier(mod)
                        break
            rgroup = level // 4
            if rgroup != group:
                kml("switching group from %i to %i", group, rgroup)
            break
        # this should not find anything new?:
        if keycode < 0:
            keycode = self.keycode_translation.get(keyname, -1)
            klog("=%i, %i (keyname translation)", keycode, rgroup)
        if keycode < 0:
            # last resort, find using the keyval:
            keymap = get_default_keymap()
            b, entries = keymap.get_entries_for_keyval(keyval)
            if b and entries:
                for entry in entries:
                    if entry.group == group:
                        return entry.keycode, entry.group
        return keycode, rgroup

    def get_current_mask(self) -> list[str]:
        # this is only called from an existing xsync context in this module
        return mask_to_names(X11Keyboard.query_mask(), self.modifier_map)

    def make_keymask_match(self, modifier_list, ignored_modifier_keycode=None, ignored_modifier_keynames=None) -> None:
        """
            Given a list of modifiers that should be set, try to press the right keys
            to make the server's modifier list match it.
            Things to take into consideration:
            * mod_managed is a list of modifiers which are "server-managed":
                these never show up in the client's modifier list as it is not aware of them,
                so we just always leave them as they are and rely on some client key event to toggle them.
                ie: "num" on win32, which is toggled by the "Num_Lock" key presses.
            * when called from '_handle_key', we ignore the modifier key which may be pressed
                or released as it should be set by that key press event.
            * when called from mouse position/click/focus events we ignore 'mod_pointermissing'
                which is set by the client to indicate modifiers which are missing from mouse events.
                ie: on win32, "lock" is missing.
                (we know this is not a keyboard event because ignored_modifier_keynames is None..)
            * if the modifier is a "nuisance" one ("lock", "num", "scroll") then we must
                simulate a full keypress (down then up).
            * some modifiers can be set by multiple keys ("shift" by both "Shift_L" and "Shift_R" for example)
                so we try to find the matching modifier in the currently pressed keys (keys_pressed)
                to make sure we unpress the right one.
        """
        if not self.keynames_for_mod:
            log("make_keymask_match: ignored as keynames_for_mod not assigned yet")
            return
        if ignored_modifier_keynames is None:
            # this is not a keyboard event, ignore modifiers in "mod_pointermissing"
            def is_ignored(modifier, _modifier_keynames) -> bool:
                return modifier in (self.mod_pointermissing or [])
        else:
            # keyboard event: ignore the keynames specified
            # (usually the modifier key being pressed/unpressed)
            def is_ignored(_modifier, modifier_keynames) -> bool:
                return len(set(modifier_keynames or []) & set(ignored_modifier_keynames or [])) > 0

        def filtered_modifiers_set(modifiers) -> set[str]:
            m: set[str] = set()
            mm = self.mod_managed or ()
            for modifier in modifiers:
                if modifier in mm:
                    log("modifier is server managed: %s", modifier)
                    continue
                keynames = self.keynames_for_mod.get(modifier)
                if is_ignored(modifier, keynames):
                    log("modifier '%s' ignored (in ignored keynames=%s)", modifier, keynames)
                    continue
                m.add(modifier)
            verboselog("filtered_modifiers_set(%s)=%s", modifiers, m)
            return m

        def change_mask(modifiers: set[str], press: bool, info: str) -> list[str]:
            failed = []
            for modifier in modifiers:
                if modifier not in self.keynames_for_mod:
                    log.error(f"Error: unknown modifier {modifier!r}")
                    log.error(" known modifiers: %s", csv(self.keynames_for_mod.keys()))
                    continue
                keynames = self.keynames_for_mod.get(modifier, ())
                if not keynames:
                    log.warn("Warning: found no keynames for '%s'", modifier)
                    continue
                # find the keycodes that match the keynames for this modifier
                keycodes = []
                # log.info("keynames(%s)=%s", modifier, keynames)
                for keyname in keynames:
                    if keyname in self.keys_pressed.values():
                        # found the key which was pressed to set this modifier
                        for keycode, name in self.keys_pressed.items():
                            if name == keyname:
                                log("found the key pressed for %s: %s", modifier, name)
                                keycodes.insert(0, keycode)
                    keycodes_for_keyname = self.keycodes_for_modifier_keynames.get(keyname, [])
                    for keycode in keycodes_for_keyname:
                        if keycode not in keycodes:
                            keycodes.append(keycode)
                if ignored_modifier_keycode is not None and ignored_modifier_keycode in keycodes:
                    log("modifier '%s' ignored (ignored keycode=%s)", modifier, ignored_modifier_keycode)
                    continue
                # nuisance keys (lock, num, scroll) are toggled by a
                # full key press + key release (so act accordingly in the loop below)
                nuisance = modifier in self.mod_nuisance
                log("keynames(%s)=%s, keycodes=%s, nuisance=%s, nuisance keys=%s",
                    modifier, keynames, keycodes, nuisance, self.mod_nuisance)
                modkeycode = None
                if not press:
                    # since we want to unpress something,
                    # let's try the keycodes we know are pressed first:
                    kdown = X11Keyboard.get_keycodes_down()
                    pressed = [x for x in keycodes if x in kdown]
                    others = [x for x in keycodes if x not in kdown]
                    keycodes = pressed + others
                for keycode in keycodes:
                    if nuisance:
                        X11Keyboard.xtest_fake_key(keycode, True)
                        X11Keyboard.xtest_fake_key(keycode, False)
                    else:
                        X11Keyboard.xtest_fake_key(keycode, press)
                    if VERIFY_MODIFIERS:
                        new_mask = self.get_current_mask()
                        success = (modifier in new_mask) == press
                    else:
                        success = True
                    if success:
                        modkeycode = keycode
                        log("change_mask(%s) %s modifier '%s' using keycode %s",
                            info, modifier_list, modifier, keycode)
                        break  # we're done for this modifier
                    log("%s %s with keycode %s did not work", info, modifier, keycode)
                    if press and not nuisance:
                        log(" trying to unpress it!")
                        X11Keyboard.xtest_fake_key(keycode, False)
                        # maybe doing the full keypress (down+up) worked:
                        new_mask = self.get_current_mask()
                        if (modifier in new_mask) == press:
                            modkeycode = keycode
                            break
                    log("change_mask(%s) %s modifier '%s' using keycode %s, success: %s",
                        info, modifier_list, modifier, keycode, success)
                if not modkeycode:
                    failed.append(modifier)
            log("change_mask(%s, %s, %s) failed=%s", modifiers, press, info, failed)
            return failed

        with xsync:
            current = filtered_modifiers_set(self.get_current_mask())
            wanted = filtered_modifiers_set(modifier_list or [])
            if current == wanted:
                return
            log("make_keymask_match(%s) current mask: %s, wanted: %s, ignoring=%s/%s, keys_pressed=%s",
                modifier_list, current, wanted,
                ignored_modifier_keycode, ignored_modifier_keynames, self.keys_pressed)
            fr = change_mask(current.difference(wanted), False, "remove")
            fa = change_mask(wanted.difference(current), True, "add")
            if not fr or fa:
                return
            if fr:
                log.warn("Warning: failed to remove the following modifiers:")
                log.warn(" %s", csv(fr))
            if fa:
                log.warn("Warning: failed to add the following modifiers:")
                log.warn(" %s", csv(fa))
            # this should never happen.. but if it does?
            # something didn't work, use the big hammer and start again from scratch:
            log.warn(" keys still pressed=%s", X11Keyboard.get_keycodes_down())
            X11Keyboard.unpress_all_keys()
            log.warn(" doing a full keyboard reset, keys now pressed=%s", X11Keyboard.get_keycodes_down())
            # and try to set the modifiers one last time:
            current = filtered_modifiers_set(self.get_current_mask())
            change_mask(current.difference(wanted), False, "remove")
            change_mask(wanted.difference(current), True, "add")



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import datetime
import platform
from typing import Any
from time import monotonic
from collections import deque
from collections.abc import Callable, Sequence

from xpra.os_util import gi_import
from xpra.util.version import XPRA_VERSION, revision_str, make_revision_str, caps_to_revision
from xpra.util.system import get_linux_distribution, platform_name
from xpra.util.objects import typedict, AtomicInteger
from xpra.util.screen import prettify_plug_name
from xpra.util.str_fn import csv, strtobytes, bytestostr, Ellipsizer
from xpra.util.env import envint
from xpra.common import noop
from xpra.util.stats import values_to_scaled_values, values_to_diff_scaled_values, to_std_unit, std_unit_dec, std_unit
from xpra.client.base import features
from xpra.client.base.command import InfoTimerClient
from xpra.gtk.window import add_close_accel
from xpra.gtk.graph import make_graph_imagesurface
from xpra.gtk.widget import imagebutton, title_box, slabel, FILE_CHOOSER_NATIVE
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.log import Logger

log = Logger("info")

GLib = gi_import("GLib")
Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")

N_SAMPLES = 20  # how many sample points to show on the graphs
SHOW_PIXEL_STATS = True
SHOW_SOUND_STATS = True
SHOW_RECV = True


def make_version_str(version) -> str:
    if version and isinstance(version, (tuple, list)):
        version = ".".join(bytestostr(x) for x in version)
    return bytestostr(version or "unknown")


def make_datetime(date, time) -> str:
    if not time:
        return bytestostr(date or "")
    return "%s %s" % (bytestostr(date), bytestostr(time))


def bool_icon(image, on_off: bool):
    if on_off:
        icon = get_icon_pixbuf("ticked-small.png")
    else:
        icon = get_icon_pixbuf("unticked-small.png")
    image.set_from_pixbuf(icon)


def pixelstr(v) -> str:
    if v < 0:
        return "n/a"
    return std_unit_dec(v)


def fpsstr(v) -> str:
    if v < 0:
        return "n/a"
    return "%s" % (int(v * 10) / 10.0)


def average(seconds, pixel_counter):
    now = monotonic()
    total = 0
    total_n = 0
    mins = None
    maxs = 0
    avgs = 0
    mint = now - seconds  # ignore records older than N seconds
    startt = now  # when we actually start counting from
    for _, t, count in pixel_counter:
        if t >= mint:
            total += count
            total_n += 1
            startt = min(t, startt)
            if mins:
                mins = min(mins, count)
            else:
                mins = count
            maxs = max(maxs, count)
            avgs += count
    if total == 0 or startt == now:
        return None
    avgs = avgs / total_n
    elapsed = now - startt
    return int(total / elapsed), total_n / elapsed, mins, avgs, maxs


def dictlook(d, k, fallback=None):
    # deal with old-style non-namespaced dicts first:
    # "batch.delay.avg"
    if d is None:
        return fallback
    v = d.get(k)
    if v is not None:
        return v
    parts = (b"client." + strtobytes(k)).split(b".")
    v = newdictlook(d, parts, fallback)
    if v is None:
        parts = strtobytes(k).split(b".")
        v = newdictlook(d, parts, fallback)
    if v is None:
        return fallback
    return v


def newdictlook(d, parts, fallback=None):
    # ie: {}, ["batch", "delay", "avg"], 0
    v = d
    for p in parts:
        try:
            newv = v.get(p)
            if newv is None:
                newv = v.get(strtobytes(p))
                if newv is None:
                    return fallback
            v = newv
        except (ValueError, TypeError, AttributeError):
            return fallback
    return v


def image_label_hbox():
    hbox = Gtk.HBox(homogeneous=False, spacing=10)
    image_widget = Gtk.Image()
    image_widget.set_size_request(48, 48)
    al_i = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0.0, yscale=0.0)
    al_i.add(image_widget)
    label_widget = slabel()
    al_l = Gtk.Alignment(xalign=0, yalign=0.5, xscale=1.0, yscale=0.0)
    al_l.add(label_widget)
    hbox.add(al_i)
    hbox.add(al_l)
    return image_widget, label_widget, hbox


def settimedeltastr(label_widget, from_time):
    import time
    delta = datetime.timedelta(seconds=(int(time.time()) - int(from_time)))
    label_widget.set_text(str(delta))


def make_os_str(sys_platform, platform_release, platform_platform, platform_linux_distribution) -> str:
    s = [platform_name(sys_platform, platform_release)]
    pld = platform_linux_distribution

    def remstr(values: Sequence) -> Sequence[str]:
        return tuple(str(v) for v in values if v not in (None, "", "n/a"))
    if pld and len(pld) == 3 and pld[0]:
        s.append(" ".join(remstr(pld)))
    elif platform_platform:
        s.append(platform_platform)
    return "\n".join(remstr(s))


def get_server_platform_name(client) -> str:
    server_info: dict = getattr(client, "server_last_info") or {}
    pinfo = server_info.get("server", {}).get("platform", {})

    def plat(key=""):
        # we can get the platform information from the server info mixin,
        # or from the server last info:
        return pinfo.get(key) or getattr(client, f"_remote_platform{key}", "")

    return make_os_str(
        plat(""),
        plat("release"),
        plat("platform"),
        plat("linux_distribution"),
    )


def get_local_platform_name() -> str:
    return make_os_str(sys.platform, platform.release(), platform.platform(), get_linux_distribution())


def get_server_builddate(client) -> str:
    build_info = client.server_last_info.get("server", {}).get("build", {})

    def cattr(name):
        return build_info.get(name, "") or getattr(client, f"_remote_build_{name}", "")

    return make_datetime(cattr("date"), cattr("time"))


def get_local_builddate() -> str:
    try:
        from xpra.build_info import build
        date = build["date"]
        time = build["time"]
        return make_datetime(date, time)
    except (ImportError, KeyError):
        return ""


def get_server_version(client) -> str:
    v = getattr(client, "_remote_version")
    return make_version_str(v or "unknown")


def get_server_revision_str(client) -> str:
    def cattr(name):
        return getattr(client, name, "")

    build_info = client.server_last_info.get("server", {}).get("build", {})
    if build_info:
        return caps_to_revision(typedict(build_info))

    return make_revision_str(
        cattr("_remote_revision"),
        cattr("_remote_modifications"),
        cattr("_remote_branch"),
        cattr("_remote_commit"),
    )


def get_glbuffering_info(opengl_props: dict) -> str:
    info = []
    tprops = typedict(opengl_props)
    display_mode = tprops.strtupleget("display_mode")
    bit_depth = tprops.intget("depth", 0)
    if bit_depth:
        info.append("%i-bit" % bit_depth)
    buffering = "unknown"
    if "double-buffered" in tprops:
        buffering = "double" if tprops.boolget("double-buffered") else "single"
    elif "DOUBLE" in display_mode:
        buffering = "double"
    elif "SINGLE" in display_mode:
        buffering = "single"
    info.append(f"{buffering} buffering")
    if tprops.intget("alpha-size", 0) > 0 or "ALPHA" in display_mode:
        info.append("with transparency")
    else:
        info.append("without transparency")
    return " ".join(info)


def set_graph_surface(graph, surface):
    w = surface.get_width()
    h = surface.get_height()
    graph.set_size_request(w, h)
    graph.surface = surface
    graph.set_from_surface(surface)


def save_graph(_ebox, btn, graph) -> None:
    log("save_graph%s", (btn, graph))
    title = "Save graph as a PNG image"
    if FILE_CHOOSER_NATIVE:
        chooser = Gtk.FileChooserNative(title=title, action=Gtk.FileChooserAction.SAVE)
        chooser.set_accept_label("Save")
    else:
        chooser = Gtk.FileChooserDialog(title=title, action=Gtk.FileChooserAction.SAVE)
        buttons = (
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
            Gtk.STOCK_SAVE, Gtk.ResponseType.OK,
        )
        chooser.add_buttons(*buttons)
        chooser.set_default_response(Gtk.ResponseType.OK)
    file_filter = Gtk.FileFilter()
    file_filter.set_name("PNG")
    file_filter.add_pattern("*.png")
    chooser.add_filter(file_filter)
    response = chooser.run()
    filenames = chooser.get_filenames()
    chooser.hide()
    if hasattr(chooser, "close"):
        chooser.close()
    if response == Gtk.ResponseType.OK:
        if len(filenames) == 1:
            filename = filenames[0]
            if not filename.lower().endswith(".png"):
                filename += ".png"
            surface = graph.surface
            log("saving surface %s to %s", surface, filename)
            from io import BytesIO
            b = BytesIO()
            surface.write_to_png(b)

            def save_file():
                with open(filename, "wb") as f:
                    f.write(b.getvalue())

            from xpra.util.thread import start_thread
            start_thread(save_file, "save-graph")
    elif response in (Gtk.ResponseType.CANCEL, Gtk.ResponseType.CLOSE, Gtk.ResponseType.DELETE_EVENT):
        log("closed/cancelled")
    else:
        log.warn(f"Warning: unknown chooser response: {response}")


class SessionInfo(Gtk.Window):

    def __init__(self, client, session_name, conn, show_client=True, show_server=True):
        assert show_client or show_server
        self.client = client
        self.session_name = session_name
        self.connection = conn
        self.show_client = show_client
        self.show_server = show_server
        super().__init__()
        self.last_populate_time = 0
        self.last_populate_statistics = 0
        self.is_closed = False
        self.set_title(self.get_window_title())
        self.set_destroy_with_parent(True)
        self.set_resizable(True)
        self.set_decorated(True)
        window_icon_pixbuf = get_icon_pixbuf("statistics.png") or get_icon_pixbuf("xpra.png")
        if window_icon_pixbuf:
            self.set_icon(window_icon_pixbuf)
        self.set_position(Gtk.WindowPosition.CENTER)

        # tab box to contain everything:
        self.tab_box = Gtk.VBox(homogeneous=False, spacing=0)
        self.add(self.tab_box)
        self.tab_button_box = Gtk.HBox(homogeneous=True, spacing=0)
        self.tabs: list[tuple[str, Any, Any, Callable]] = []
        self.row = AtomicInteger()
        self.grid = None

        self.populate_cb = None
        self.tab_box.pack_start(self.tab_button_box, expand=False, fill=True, padding=0)

        self.add_software_tab()
        self.add_features_tab()
        self.add_codecs_tab()
        self.add_connection_tab()
        self.add_statistics_tab()
        if show_client:
            self.add_graphs_tab()

        self.set_border_width(15)

        def window_deleted(*_args):
            self.is_closed = True

        self.connect('delete_event', window_deleted)
        self.show_tab(self.tabs[0][2])
        self.set_size_request(-1, -1)
        self.init_counters()
        self.populate()
        self.populate_all()
        GLib.timeout_add(1000, self.populate)
        GLib.timeout_add(100, self.populate_tab)
        if features.audio and SHOW_SOUND_STATS and show_client:
            GLib.timeout_add(100, self.populate_audio_stats)
        add_close_accel(self, self.close)

    def get_window_title(self) -> str:
        t = ["Session Info"]
        c = self.client
        if c:
            if c.session_name or c.server_session_name:
                t.append(c.session_name or c.server_session_name)
            p = c._protocol
            if p:
                conn = getattr(p, "_conn", None)
                if conn:
                    cinfo = conn.get_info()
                    t.append(cinfo.get("endpoint", bytestostr(conn.target)))
        v = " - ".join(str(x) for x in t)
        return v

    def add_row(self, *widgets):
        r = int(self.row)
        for i, widget in enumerate(widgets):
            self.grid.attach(widget, i, r, 1, 1)
        self.row.increase()

    def new_row(self, text: str, *widgets):
        self.add_row(slabel(text), *widgets)

    def label_row(self, text: str, value=""):
        lbl = slabel(value)
        al = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0.0, yscale=0.0)
        al.add(lbl)
        self.add_row(title_box(text), al)
        return lbl

    def clrow(self, label_str: str, client_label, server_label):
        labels = []
        if self.show_client:
            labels.append(client_label)
        if self.show_server:
            labels.append(server_label)
        self.add_row(title_box(label_str), *labels)

    def csrow(self, label_str, client_str, server_str):
        self.clrow(
            label_str,
            slabel(client_str) if self.show_client else None,
            slabel(server_str) if self.show_server else None,
        )

    def grid_tab(self, icon_filename: str, title: str, populate_cb: Callable):
        self.grid = Gtk.Grid()
        self.row.set(0)
        vbox = self.vbox_tab(icon_filename, title, populate_cb)
        al = Gtk.Alignment(xalign=0.5, yalign=0.5, xscale=0.0, yscale=1.0)
        al.add(self.grid)
        vbox.pack_start(al, expand=True, fill=True, padding=20)
        return vbox

    def vbox_tab(self, icon_filename: str, title: str, populate_cb: Callable):
        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        self.add_tab(icon_filename, title, populate_cb, contents=vbox)
        return vbox

    def add_tab(self, icon_filename: str, title: str, populate_cb: Callable, contents):
        icon = get_icon_pixbuf(icon_filename)

        def show_tab(*_args):
            self.show_tab(contents)

        button = imagebutton(title, icon, clicked_callback=show_tab)
        button.connect("clicked", show_tab)
        button.set_relief(Gtk.ReliefStyle.NONE)
        self.tab_button_box.add(button)
        self.tabs.append((title, button, contents, populate_cb))

    def show_tab(self, grid):
        button = None
        for _, b, t, p_cb in self.tabs:
            if t == grid:
                button = b
                b.set_relief(Gtk.ReliefStyle.NORMAL)
                b.grab_focus()
                self.populate_cb = p_cb
            else:
                b.set_relief(Gtk.ReliefStyle.NONE)
        assert button
        for x in self.tab_box.get_children():
            if x != self.tab_button_box:
                self.tab_box.remove(x)
        self.tab_box.pack_start(grid, expand=True, fill=True, padding=0)
        grid.show_all()
        # ensure we re-draw the whole window:
        window = self.get_window()
        if window:
            alloc = self.get_allocation()
            window.invalidate_rect(alloc, True)

    def set_args(self, *args):
        # this is a generic way for keyboard shortcuts or remote commands
        # to pass parameters to us
        log("set_args%s", args)
        if not args:
            return
        # at the moment, we only handle the tab name as argument:
        tab_name = args[0]
        if tab_name.lower() != "help":
            title = ""
            for title, _, table, _ in self.tabs:
                if title.lower() == tab_name.lower():
                    self.show_tab(table)
                    return
            log.warn("could not find session info tab named: %s", title)
        log.warn("The options for tab names are: %s)", [x[0] for x in self.tabs])

    def populate_all(self):
        for tab in self.tabs:
            p_cb = tab[3]
            if p_cb:
                p_cb()

    def add_graph_button(self, tooltip: str = "", click_cb: Callable = noop) -> Gtk.Image:
        button = Gtk.EventBox()
        try:
            display = Gdk.Display.get_default()
            arrow_down = Gdk.Cursor.new_for_display(display, Gdk.CursorType.BASED_ARROW_DOWN)
        except TypeError:
            arrow_down = None

        def set_cursor(widget) -> None:
            widget.get_window().set_cursor(arrow_down)

        if arrow_down:
            button.connect("realize", set_cursor)
        graph = Gtk.Image()
        graph.set_size_request(0, 0)
        button.connect("button_press_event", click_cb, graph)
        button.add(graph)
        if tooltip:
            graph.set_tooltip_text(tooltip)
        self.graph_box.add(button)
        return graph

    def populate_audio_stats(self, *_args):
        # runs every 100ms
        if self.is_closed:
            return False
        ss = self.client.audio_sink
        if SHOW_SOUND_STATS and ss:
            info = ss.get_info()
            if info:
                info = typedict(info)

                def qlookup(attr) -> int:
                    return int(newdictlook(info, ("queue", attr), 0))

                self.audio_out_queue_cur.append(qlookup("cur"))
                self.audio_out_queue_min.append(qlookup("min"))
                self.audio_out_queue_max.append(qlookup("max"))
        return not self.is_closed

    def populate(self, *_args):
        conn = self.connection
        if self.is_closed or not conn:
            return False
        self.client.send_ping()
        self.last_populate_time = monotonic()

        if self.show_client:
            self.show_opengl_state()
            self.show_window_renderers()
            # record bytecount every second:
            self.net_in_bitcount.append(conn.input_bytecount * 8)
            self.net_out_bitcount.append(conn.output_bytecount * 8)
            if features.audio and SHOW_SOUND_STATS:
                if self.client.audio_in_bytecount > 0:
                    self.audio_in_bitcount.append(self.client.audio_in_bytecount * 8)
                if self.client.audio_out_bytecount > 0:
                    self.audio_out_bitcount.append(self.client.audio_out_bytecount * 8)

        if self.show_client and features.windows:
            # count pixels in the last second:
            since = monotonic() - 1
            decoded = [0] + [pixels for _, t, pixels in self.client.pixel_counter if t > since]
            self.pixel_in_data.append(sum(decoded))

        # update latency values
        # there may be more than one record for each second,
        # so we have to average them to prevent the graph from "jumping":

        def get_ping_latency_records(src, size=25):
            recs = {}
            src_list = list(src)
            now = int(monotonic())
            while src_list and len(recs) < size:
                when, value = src_list.pop()
                if when >= (now - 1):  # ignore last second
                    continue
                iwhen = int(when)
                cv = recs.get(iwhen)
                v = 1000.0 * value
                if cv:
                    v = (v + cv) / 2.0  # not very fair if more than 2 values... but this shouldn't happen anyway
                recs[iwhen] = v
            # ensure we always have a record for the last N seconds, even an empty one
            for x in range(size):
                i = now - 2 - x
                if i not in recs:
                    recs[i] = None
            return tuple(recs.get(x) for x in sorted(recs.keys()))

        self.server_latency = get_ping_latency_records(self.client.server_ping_latency)
        self.client_latency = get_ping_latency_records(self.client.client_ping_latency)
        if self.client.server_last_info:
            # populate running averages for graphs:

            def getavg(*names):
                return newdictlook(self.client.server_last_info, list(names) + ["avg"]) or \
                    newdictlook(self.client.server_last_info, ["client"] + list(names) + ["avg"])

            def addavg(l, *names):
                v = getavg(*names)
                if v:
                    l.append(v)

            addavg(self.avg_batch_delay, "batch", "delay")
            addavg(self.avg_damage_out_latency, "damage", "out_latency")
            spl = tuple(1000.0 * x[1] for x in tuple(self.client.server_ping_latency))
            cpl = tuple(1000.0 * x[1] for x in tuple(self.client.client_ping_latency))
            if spl and cpl:
                self.avg_ping_latency.append(round(sum(spl + cpl) / len(spl + cpl)))
            if features.windows and self.show_client:
                pc = tuple(self.client.pixel_counter)
                if pc:
                    tsize = 0
                    ttime = 0
                    for start_time, end_time, size in pc:
                        ttime += 1000.0 * (end_time - start_time) * size
                        tsize += size
                    self.avg_decoding_latency.append(round(ttime / tsize))
        # totals: ping latency is halved since we only care about sending, not sending+receiving
        els = (
            (tuple(self.avg_batch_delay), 1),
            (tuple(self.avg_damage_out_latency), 1),
            (tuple(self.avg_ping_latency), 2),
            (tuple(self.avg_decoding_latency), 1),
        )
        if all(x[0] for x in els):
            totals = tuple(x[-1] / r for x, r in els)
            log("frame totals=%s", totals)
            self.avg_total.append(round(sum(totals)))
        return not self.is_closed

    def add_software_tab(self):
        # Package Table:
        self.grid_tab("package.png", "Software", noop)
        if self.show_client and self.show_server:
            self.add_row(title_box(""), title_box("Client"), title_box("Server"))
        self.csrow("Operating System", get_local_platform_name(), get_server_platform_name(self.client))
        self.csrow("Xpra", XPRA_VERSION, get_server_version(self.client))
        self.csrow("Revision", revision_str(), get_server_revision_str(self.client))
        self.csrow("Build date", get_local_builddate(), get_server_builddate(self.client))

        def server_vinfo(lib):
            rlv = getattr(self.client, "_remote_lib_versions", {})
            return make_version_str(rlv.get(lib, ""))

        try:
            from xpra.audio.wrapper import query_audio
            props = query_audio()
        except ImportError:
            log("cannot load audio information: %s", exc_info=True)
            props = typedict()
        gst_version = props.strtupleget("gst.version")
        self.csrow("GStreamer", make_version_str(gst_version), server_vinfo("sound.gst") or server_vinfo("gst") or "unknown")

        def clientgl(prop="opengl", default_value="n/a"):
            if not self.show_client:
                return ""
            return make_version_str(self.client.opengl_props.get(prop, default_value))

        def servergl(prop="opengl", default_value="n/a"):
            if not self.show_server:
                return ""
            return make_version_str(typedict(self.client.server_opengl or {}).get(prop, default_value))

        for prop in ("OpenGL", "Vendor", "PyOpenGL"):
            key = prop.lower()
            self.csrow(prop, clientgl(key), servergl(key))

    def init_counters(self):
        self.avg_batch_delay = deque(maxlen=N_SAMPLES + 4)
        self.avg_damage_out_latency = deque(maxlen=N_SAMPLES + 4)
        self.avg_ping_latency = deque(maxlen=N_SAMPLES + 4)
        self.avg_decoding_latency = deque(maxlen=N_SAMPLES + 4)
        self.avg_total = deque(maxlen=N_SAMPLES + 4)

    def populate_tab(self, *_args):
        if self.is_closed:
            return False
        # now re-populate the tab we are seeing:
        if self.populate_cb and not self.populate_cb():
            self.populate_cb = None
        return not self.is_closed

    def show_opengl_state(self):
        if self.client.opengl_enabled:
            glinfo = "%s / %s" % (
                self.client.opengl_props.get("vendor", ""),
                self.client.opengl_props.get("renderer", ""),
            )
            info = get_glbuffering_info(self.client.opengl_props)
        else:
            # info could be telling us that the gl bindings are missing:
            glinfo = self.client.opengl_props.get("info", "disabled")
            info = "n/a"
        self.client_opengl_label.set_text(glinfo)
        self.opengl_buffering.set_text(info)

    def show_window_renderers(self):
        if not features.windows:
            return
        wr = []
        renderers = {}
        for wid, window in tuple(self.client._id_to_window.items()):
            renderers.setdefault(window.get_backing_class(), []).append(wid)
        for bclass, windows in renderers.items():
            wr.append("%s (%i)" % (bclass.__name__.replace("Backing", ""), len(windows)))
        self.window_rendering.set_text("GTK3: %s" % csv(wr))

    def add_features_tab(self):
        # Features Table:
        self.grid_tab("features.png", "Features", self.populate_features)

        def image_row(text: str):
            img = Gtk.Image()
            img.set_margin_start(5)
            al = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0.0, yscale=0.0)
            al.add(img)
            self.add_row(title_box(text), al)
            return img

        self.server_randr_icon, self.server_randr_label, randr_box = image_label_hbox()
        self.add_row(title_box("RandR Support"), randr_box)
        if self.show_client:
            self.client_display = self.label_row("Client Display")
            self.client_opengl_icon, self.client_opengl_label, opengl_box = image_label_hbox()
            self.add_row(title_box("Client OpenGL"), opengl_box)
            self.opengl_buffering = self.label_row("OpenGL Mode")
            self.window_rendering = self.label_row("Window Rendering")
        if features.mmap:
            self.server_mmap_icon = image_row("Memory Mapped Transfers")
        if features.clipboard:
            self.server_clipboard_icon = image_row("Clipboard")
        if features.notifications:
            self.server_notifications_icon = image_row("Notifications")
        if features.windows:
            self.server_bell_icon = image_row("Bell")
            self.server_cursors_icon = image_row("Cursors")

    def populate_features(self):
        size_info = ""
        if features.windows:
            if self.client.server_actual_desktop_size:
                w, h = self.client.server_actual_desktop_size
                size_info = f"{w}x{h}"
                if self.client.server_randr and self.client.server_max_desktop_size:
                    size_info += " (max %s)" % ("x".join([str(x) for x in self.client.server_max_desktop_size]))
                bool_icon(self.server_randr_icon, self.client.server_randr)
        else:
            size_info = "unknown"
            unknown = get_icon_pixbuf("unknown.png")
            if unknown:
                self.server_randr_icon.set_from_pixbuf(unknown)
        self.server_randr_label.set_text("%s" % size_info)
        if self.show_client:
            root_w, root_h = self.client.get_root_size()
            if features.windows and (self.client.xscale != 1 or self.client.yscale != 1):
                sw, sh = self.client.cp(root_w, root_h)
                display_info = "%ix%i (scaled from %ix%i)" % (sw, sh, root_w, root_h)
            else:
                display_info = "%ix%i" % (root_w, root_h)
            self.client_display.set_text(display_info)
            bool_icon(self.client_opengl_icon, self.client.client_supports_opengl)
            self.show_window_renderers()

        if features.mmap:
            bool_icon(self.server_mmap_icon, bool(self.client.mmap_read_area))
        if features.clipboard:
            bool_icon(self.server_clipboard_icon, self.client.server_clipboard)
        if features.notifications:
            bool_icon(self.server_notifications_icon, self.client.server_notifications)
        if features.windows:
            bool_icon(self.server_bell_icon, self.client.server_bell)
            bool_icon(self.server_cursors_icon, self.client.server_cursors)

    def add_codecs_tab(self):
        # Codecs Table:
        self.grid_tab("encoding.png", "Codecs", self.populate_codecs)
        if self.show_client and self.show_server:
            # table headings:
            for i, text in enumerate(("", "Client", "Server")):
                self.grid.attach(title_box(text), i, 0, 1, 1)
            self.row.increase()
        # grid contents:
        self.client_encodings_label = slabel()
        self.server_encodings_label = slabel()
        self.clrow("Picture Encodings", self.client_encodings_label, self.server_encodings_label)
        self.client_speaker_codecs_label = slabel()
        self.server_speaker_codecs_label = slabel()
        self.clrow("Speaker Codecs", self.client_speaker_codecs_label, self.server_speaker_codecs_label)
        self.client_microphone_codecs_label = slabel()
        self.server_microphone_codecs_label = slabel()
        self.clrow("Microphone Codecs", self.client_microphone_codecs_label, self.server_microphone_codecs_label)
        self.client_packet_encoders_label = slabel()
        self.server_packet_encoders_label = slabel()
        self.clrow("Packet Encoders", self.client_packet_encoders_label, self.server_packet_encoders_label)
        self.client_packet_compressors_label = slabel()
        self.server_packet_compressors_label = slabel()
        self.clrow("Packet Compressors", self.client_packet_compressors_label, self.server_packet_compressors_label)

    def populate_codecs(self):
        # clamp the large labels so they will overflow vertically:
        w = self.tab_box.get_preferred_width()[0]
        lw = max(200, int(w // 2.5))
        if self.show_client:
            self.client_encodings_label.set_size_request(lw, -1)
            self.client_speaker_codecs_label.set_size_request(lw, -1)
            self.client_microphone_codecs_label.set_size_request(lw, -1)
        if self.show_server:
            self.server_encodings_label.set_size_request(lw, -1)
            self.server_speaker_codecs_label.set_size_request(lw, -1)
            self.server_microphone_codecs_label.set_size_request(lw, -1)

        # audio/video codec table:

        def codec_info(enabled, codecs):
            if not enabled:
                return "n/a"
            return ", ".join(codecs or ())

        if features.audio:
            c = self.client
            if self.show_server:
                self.server_speaker_codecs_label.set_text(
                    codec_info(c.server_audio_send, c.server_audio_encoders))
            if self.show_client:
                self.client_speaker_codecs_label.set_text(
                    codec_info(c.speaker_allowed, c.speaker_codecs))
            if self.show_server:
                self.server_microphone_codecs_label.set_text(
                    codec_info(c.server_audio_receive, c.server_audio_decoders))
            if self.show_client:
                self.client_microphone_codecs_label.set_text(
                    codec_info(c.microphone_allowed, c.microphone_codecs))

        def encliststr(v):
            v = list(v)
            try:
                v.remove("rgb")
            except ValueError:
                pass
            return csv(sorted(v))

        se = ()
        if features.encoding:
            se = self.client.server_core_encodings
        self.server_encodings_label.set_text(encliststr(se))
        if self.show_client and features.encoding:
            self.client_encodings_label.set_text(encliststr(self.client.get_core_encodings()))
        else:
            self.client_encodings_label.set_text("n/a")

        from xpra.net.packet_encoding import get_enabled_encoders
        if self.show_client:
            self.client_packet_encoders_label.set_text(csv(get_enabled_encoders()))
        if self.show_server:
            self.server_packet_encoders_label.set_text(csv(self.client.server_packet_encoders))

        from xpra.net.compression import get_enabled_compressors
        if self.show_client:
            self.client_packet_compressors_label.set_text(csv(get_enabled_compressors()))
        if self.show_server:
            self.server_packet_compressors_label.set_text(csv(self.client.server_compressors))
        return False

    def add_connection_tab(self):
        self.grid_tab("connect.png", "Connection", self.populate_connection)

        def cattr(name) -> str:
            return getattr(self.client, name, "")

        if self.connection:
            self.connection.target = self.label_row("Server Endpoint")
        if features.display and self.client.server_display:
            self.label_row("Server Display", prettify_plug_name(self.client.server_display))
        self.label_row("Server Hostname", cattr("_remote_hostname"))
        if cattr("server_platform"):
            self.label_row("Server Platform", cattr("server_platform"))
        self.server_load_label = self.label_row("Server Load")
        self.server_load_label.set_tooltip_text("Average over 1, 5 and 15 minutes")
        self.session_started_label = self.label_row("Session Started")
        if not self.show_client:
            return
        self.session_connected_label = self.label_row("Session Connected")
        self.input_packets_label = self.label_row("Packets Received")
        self.input_bytes_label = self.label_row("Bytes Received")
        self.output_packets_label = self.label_row("Packets Sent")
        self.output_bytes_label = self.label_row("Bytes Sent")
        self.compression_label = self.label_row("Encoding + Compression")
        self.connection_type_label = self.label_row("Connection Type")
        self.input_encryption_label = self.label_row("Input Encryption")
        self.output_encryption_label = self.label_row("Output Encryption")

        def add_audio_row(text):
            lbl = slabel()
            al = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0.0, yscale=0.0)
            al.add(lbl)
            details = slabel(font="monospace 10")
            self.add_row(title_box(text), al, details)
            return lbl, details

        self.speaker_label, self.speaker_details = add_audio_row("Speaker")
        self.microphone_label, self.microphone_details = add_audio_row("Microphone")

    def populate_connection(self):
        if self.client.server_load:
            self.server_load_label.set_text("  ".join("%.1f" % (x / 1000.0) for x in self.client.server_load))
        if self.client.server_start_time > 0:
            settimedeltastr(self.session_started_label, self.client.server_start_time)
        else:
            self.session_started_label.set_text("unknown")
        settimedeltastr(self.session_connected_label, self.client.start_time)

        p = self.client._protocol
        if not self.show_client:
            return True
        if p is None:
            # no longer connected!
            return False
        c = p._conn
        if c:
            self.input_packets_label.set_text(std_unit_dec(p.input_packetcount))
            self.input_bytes_label.set_text(std_unit_dec(c.input_bytecount))
            self.output_packets_label.set_text(std_unit_dec(p.output_packetcount))
            self.output_bytes_label.set_text(std_unit_dec(c.output_bytecount))
        else:
            for lbl in (
                    self.input_packets_label,
                    self.input_bytes_label,
                    self.output_packets_label,
                    self.output_bytes_label,
            ):
                lbl.set_text("n/a")

        if features.audio:

            def get_audio_info(supported, prop):
                if not supported:
                    return {"state": "disabled"}
                if prop is None:
                    return {"state": "inactive"}
                return prop.get_info()

            def set_audio_info(label, details, supported, prop):
                d = typedict(get_audio_info(supported, prop))
                state = d.strget("state")
                codec_descr = d.strget("codec") or d.strget("codec_description")
                container_descr = d.strget("container_description")
                if state == "active" and codec_descr:
                    if codec_descr.find(container_descr) >= 0:
                        descr = codec_descr
                    else:
                        descr = csv(x for x in (codec_descr, container_descr) if x)
                    state = f"{state}: {descr}"
                label.set_text(state)
                if details:
                    s = ""
                    bitrate = d.intget("bitrate", 0)
                    if bitrate > 0:
                        s = "%sbit/s" % std_unit(bitrate)
                    details.set_text(s)

            set_audio_info(self.speaker_label, self.speaker_details,
                           self.client.speaker_enabled, self.client.audio_sink)
            set_audio_info(self.microphone_label, None,
                           self.client.microphone_enabled, self.client.audio_source)

        self.connection_type_label.set_text(c.socktype)
        protocol_info = p.get_info()
        encoder = protocol_info.get("encoder", "bug")
        compressor = protocol_info.get("compressor", "none")
        level = protocol_info.get("compression_level", 0)
        compression_str = encoder + " + " + compressor
        if level > 0:
            compression_str += " (level %s)" % level
        self.compression_label.set_text(compression_str)

        from xpra.net.crypto import get_crypto_caps
        ccaps = get_crypto_caps()

        def enclabel(label_widget, cipher):
            if not cipher:
                info = "None"
            else:
                info = str(cipher)
            if c.socktype.lower() == "ssh":
                info += " (%s)" % c.socktype
            backend = ccaps.get("backend")
            if backend == "python-cryptography":
                info += " / python-cryptography"
            label_widget.set_text(info)

        enclabel(self.input_encryption_label, p.cipher_in_name)
        enclabel(self.output_encryption_label, p.cipher_out_name)
        return True

    def getval(self, prefix, suffix, alt=""):
        if self.client.server_last_info is None:
            return ""
        altv = ""
        if alt:
            altv = dictlook(self.client.server_last_info, (alt + "." + suffix).encode(), "")
        return dictlook(self.client.server_last_info, (prefix + "." + suffix).encode(), altv)

    def values_from_info(self, prefix, alt=None):
        def getv(suffix):
            return self.getval(prefix, suffix, alt)

        return getv("cur"), getv("min"), getv("avg"), getv("90p"), getv("max")

    def all_values_from_info(self, *window_props):
        for window_prop in window_props:
            prop_path = "client.%s" % window_prop
            v = dictlook(self.client.server_last_info, prop_path)
            if v is not None:
                v = typedict(v)
                iget = v.intget
                return iget("cur"), iget("min"), iget("avg"), iget("90p"), iget("max")
        return 0, 0, 0, 0, 0

    def add_statistics_tab(self):
        # Details:
        vbox = self.grid_tab("browse.png", "Statistics", self.populate_statistics)
        self.add_row(*(title_box(x) for x in ("", "Latest", "Minimum", "Average", "90 percentile", "Maximum")))

        def maths_labels(metric="", tooltip=""):
            labels = [title_box(metric, tooltip), slabel(), slabel(), slabel(), slabel(), slabel()]
            self.add_row(*labels)
            return labels[1:]

        self.server_latency_labels = maths_labels(
            "Server Latency (ms)",
            "The time it takes for the server to respond to pings",
        )
        self.client_latency_labels = maths_labels(
            "Client Latency (ms)",
            "The time it takes for the client to respond to pings, as measured by the server",
        )
        if not features.windows or not self.client.windows_enabled:
            return
        self.batch_labels = maths_labels(
            "Batch Delay (MPixels / ms)",
            "How long the server waits for new screen updates to accumulate before processing them",
        )
        self.damage_labels = maths_labels(
            "Damage Latency (ms)",
            "The time it takes to compress a frame and pass it to the OS network layer",
        )
        self.quality_labels = maths_labels(
            "Encoding Quality (pct)",
            "Automatic picture quality, average for all the windows"
        )
        self.speed_labels = maths_labels(
            "Encoding Speed (pct)",
            "Automatic picture encoding speed (bandwidth vs CPU usage), average for all the windows",
        )
        self.decoding_labels = maths_labels(
            "Decoding Latency (ms)",
            "How long it takes the client to decode a screen update",
        )
        self.regions_per_second_labels = maths_labels(
            "Regions/s",
            "The number of screen updates per second (includes both partial and full screen updates)",
        )
        self.regions_sizes_labels = maths_labels(
            "Pixels/region",
            "The number of pixels per screen update",
        )
        self.pixels_per_second_labels = maths_labels(
            "Pixels/s",
            "The number of pixels processed per second",
        )

        # grid 2:
        self.grid = Gtk.Grid()
        self.row.set(0)
        vbox.add(self.grid)
        self.add_row(*(title_box(x) for x in ("", "Regular", "Transient", "Trays", "OpenGL")))
        self.windows_managed_label = slabel()
        self.transient_managed_label = slabel()
        self.trays_managed_label = slabel()
        self.opengl_label = slabel()
        self.add_row(title_box("Windows"),
                     self.windows_managed_label, self.transient_managed_label,
                     self.trays_managed_label, self.opengl_label)

        self.encoder_info_box = Gtk.HBox(spacing=4)
        self.encoder_info_box.add(title_box("Window Encoders"))
        al = Gtk.Alignment(xalign=0.5, yalign=0.5, xscale=1.0, yscale=0.0)
        al.set_margin_start(10)
        al.set_margin_end(10)
        al.add(self.encoder_info_box)
        vbox.add(al)

    def populate_statistics(self):
        log("populate_statistics()")
        if monotonic() - self.last_populate_statistics < 1.0:
            # don't repopulate more than every second
            return True
        self.last_populate_statistics = monotonic()
        self.client.send_info_request()

        def setall(labels, values):
            assert len(labels) == len(values), "%s labels and %s values (%s vs %s)" % (
                len(labels), len(values), labels, values)
            for i, l in enumerate(labels):
                l.set_text(str(values[i]))

        def setlabels(labels, values, rounding: Callable = int):
            if not values:
                return
            avg = sum(values) / len(values)
            svalues = sorted(values)
            l = len(svalues)
            assert l > 0
            if l < 10:
                index = l - 1
            else:
                index = int(l * 90 / 100)
            index = max(0, min(l - 1, index))
            pct = svalues[index]
            disp = values[-1], min(values), avg, pct, max(values)
            rounded_values = [rounding(v) for v in disp]
            setall(labels, rounded_values)

        if self.client.server_ping_latency:
            spl = tuple(int(1000 * x[1]) for x in tuple(self.client.server_ping_latency))
            setlabels(self.server_latency_labels, spl)
        if self.client.client_ping_latency:
            cpl = tuple(int(1000 * x[1]) for x in tuple(self.client.client_ping_latency))
            setlabels(self.client_latency_labels, cpl)
        if features.windows and self.client.windows_enabled:
            setall(self.batch_labels, self.values_from_info("batch_delay", "batch.delay"))
            setall(self.damage_labels, self.values_from_info("damage_out_latency", "damage.out_latency"))
            setall(self.quality_labels, self.all_values_from_info("quality", "encoding.quality"))
            setall(self.speed_labels, self.all_values_from_info("speed", "encoding.speed"))

            region_sizes = []
            rps = []
            pps = []
            decoding_latency = []
            if self.client.pixel_counter:
                min_time = None
                max_time = None
                regions_per_second = {}
                pixels_per_second = {}
                for start_time, end_time, size in self.client.pixel_counter:
                    decoding_latency.append(int(1000.0 * (end_time - start_time)))
                    region_sizes.append(size)
                    if min_time is None or min_time > end_time:
                        min_time = end_time
                    if max_time is None or max_time < end_time:
                        max_time = end_time
                    time_in_seconds = int(end_time)
                    regions = regions_per_second.get(time_in_seconds, 0)
                    regions_per_second[time_in_seconds] = regions + 1
                    pixels = pixels_per_second.get(time_in_seconds, 0)
                    pixels_per_second[time_in_seconds] = pixels + size
                if int(min_time) + 1 < int(max_time):
                    for t in range(int(min_time) + 1, int(max_time)):
                        rps.append(regions_per_second.get(t, 0))
                        pps.append(pixels_per_second.get(t, 0))
            setlabels(self.decoding_labels, decoding_latency)
            setlabels(self.regions_per_second_labels, rps)
            setlabels(self.regions_sizes_labels, region_sizes, rounding=std_unit_dec)
            setlabels(self.pixels_per_second_labels, pps, rounding=std_unit_dec)

            windows, gl, transient, trays = 0, 0, 0, 0
            for w in self.client._window_to_id.keys():
                if w.is_tray():
                    trays += 1
                elif w.is_OR():
                    transient += 1
                else:
                    windows += 1
                if w.is_GL():
                    gl += 1
            self.windows_managed_label.set_text(str(windows))
            self.transient_managed_label.set_text(str(transient))
            self.trays_managed_label.set_text(str(trays))
            if self.client.client_supports_opengl:
                self.opengl_label.set_text(str(gl))

            # remove all the current labels:
            for x in self.encoder_info_box.get_children():
                self.encoder_info_box.remove(x)
            if self.client.server_last_info:
                window_encoder_stats = self.get_window_encoder_stats()
                # log("window_encoder_stats=%s", window_encoder_stats)
                for wid, props in window_encoder_stats.items():
                    lbl = slabel("%s (%s)" % (wid, bytestostr(props.get(""))))
                    lbl.show()
                    info = ("%s=%s" % (k, v) for k, v in props.items() if k != "")
                    lbl.set_tooltip_text(" ".join(info))
                    self.encoder_info_box.add(lbl)
        return True

    def get_window_encoder_stats(self):
        window_encoder_stats = {}
        # new-style server with namespace (easier):
        window_dict = self.client.server_last_info.get("window")
        if window_dict and isinstance(window_dict, dict):
            for k, v in window_dict.items():
                with log.trap_error("Error: cannot lookup window dict"):
                    wid = int(k)
                    encoder_stats = v.get("encoder")
                    if encoder_stats:
                        window_encoder_stats[wid] = encoder_stats
        return window_encoder_stats

    def add_graphs_tab(self):
        self.graph_box = Gtk.VBox(homogeneous=False, spacing=10)
        self.add_tab("statistics.png", "Graphs", self.populate_graphs, self.graph_box)
        bandwidth_label = "Bandwidth used"
        if SHOW_PIXEL_STATS:
            bandwidth_label += ",\nand number of pixels rendered"
        self.bandwidth_graph = self.add_graph_button(bandwidth_label, save_graph)
        self.latency_graph = self.add_graph_button("", save_graph)
        if SHOW_SOUND_STATS:
            self.audio_queue_graph = self.add_graph_button("", save_graph)
        else:
            self.audio_queue_graph = None
        self.connect("realize", self.populate_graphs)
        self.pixel_in_data = deque(maxlen=N_SAMPLES + 4)
        self.net_in_bitcount = deque(maxlen=N_SAMPLES + 4)
        self.net_out_bitcount = deque(maxlen=N_SAMPLES + 4)
        self.audio_in_bitcount = deque(maxlen=N_SAMPLES + 4)
        self.audio_out_bitcount = deque(maxlen=N_SAMPLES + 4)
        self.audio_out_queue_min = deque(maxlen=N_SAMPLES * 10 + 4)
        self.audio_out_queue_max = deque(maxlen=N_SAMPLES * 10 + 4)
        self.audio_out_queue_cur = deque(maxlen=N_SAMPLES * 10 + 4)

    def populate_graphs(self, *_args):
        # older servers have 'batch' at top level,
        # newer servers store it under client
        self.client.send_info_request("network", "damage", "state", "batch", "client")
        box = self.tab_box
        h = box.get_preferred_height()[0]
        bh = self.tab_button_box.get_preferred_height()[0]
        if h <= 0:
            return True
        start_x_offset = min(1.0, (monotonic() - self.last_populate_time) * 0.95)
        rect = box.get_allocation()
        maxw, maxh = self.client.get_root_size()
        ngraphs = 2 + int(SHOW_SOUND_STATS)
        # the preferred size (which does not cause the window to grow too big):
        W = 360
        H = 160 * 3 // ngraphs
        w = min(maxw, max(W, rect.width - 20))
        # need some padding to be able to shrink the window again:
        pad = 50
        h = min(maxh - pad // ngraphs, max(H, (h - bh - pad) // ngraphs, (rect.height - bh - pad) // ngraphs))
        # bandwidth graph:
        labels, datasets = [], []

        def unit(scale):
            if scale == 1:
                return ""
            unit, value = to_std_unit(scale)
            if value == 1:
                return str(unit)
            return f"x{int(value)}{unit}"

        if self.net_in_bitcount and self.net_out_bitcount:
            net_in_scale, net_in_data = values_to_diff_scaled_values(tuple(
                self.net_in_bitcount)[1:N_SAMPLES + 3], scale_unit=1000, min_scaled_value=50)
            net_out_scale, net_out_data = values_to_diff_scaled_values(tuple(
                self.net_out_bitcount)[1:N_SAMPLES + 3], scale_unit=1000, min_scaled_value=50)
            if SHOW_RECV:
                labels += ["recv %sb/s" % unit(net_in_scale), "sent %sb/s" % unit(net_out_scale)]
                datasets += [net_in_data, net_out_data]
            else:
                labels += ["recv %sb/s" % unit(net_in_scale)]
                datasets += [net_in_data]
        if features.windows and SHOW_PIXEL_STATS and self.client.windows_enabled:
            pixel_scale, in_pixels = values_to_scaled_values(tuple(
                self.pixel_in_data)[3:N_SAMPLES + 4], min_scaled_value=100)
            datasets.append(in_pixels)
            labels.append("%s pixels/s" % unit(pixel_scale))
        if features.audio and SHOW_SOUND_STATS and self.audio_in_bitcount:
            audio_in_scale, audio_in_data = values_to_diff_scaled_values(tuple(
                self.audio_in_bitcount)[1:N_SAMPLES + 3], scale_unit=1000, min_scaled_value=50)
            datasets.append(audio_in_data)
            labels.append("Speaker %sb/s" % unit(audio_in_scale))
        if features.audio and SHOW_SOUND_STATS and self.audio_out_bitcount:
            audio_out_scale, audio_out_data = values_to_diff_scaled_values(tuple(
                self.audio_out_bitcount)[1:N_SAMPLES + 3], scale_unit=1000, min_scaled_value=50)
            datasets.append(audio_out_data)
            labels.append("Mic %sb/s" % unit(audio_out_scale))

        if labels and datasets:
            surface = make_graph_imagesurface(datasets, labels=labels,
                                              width=w, height=h,
                                              title="Bandwidth", min_y_scale=10, rounding=10,
                                              start_x_offset=start_x_offset)
            set_graph_surface(self.bandwidth_graph, surface)

        def norm_lists(items, size=N_SAMPLES):
            # ensures we always have exactly 20 values,
            # (and skip if we don't have any)
            values, labels = [], []
            for l, name in items:
                if not l:
                    continue
                l = list(l)
                if len(l) < size:
                    for _ in range(size - len(l)):
                        l.insert(0, None)
                else:
                    l = l[:size]
                values.append(l)
                labels.append(name)
            return values, labels

        # latency graph:
        latency_values, latency_labels = norm_lists(
            (
                (self.avg_ping_latency, "network"),
                (self.avg_batch_delay, "batch delay"),
                (self.avg_damage_out_latency, "encode&send"),
                (self.avg_decoding_latency, "decoding"),
                (self.avg_total, "frame total"),
            )
        )
        # debug:
        # for i, v in enumerate(latency_values):
        #    log.warn("%20s = %s", latency_labels[i], v)
        surface = make_graph_imagesurface(latency_values, labels=latency_labels,
                                          width=w, height=h,
                                          title="Latency (ms)", min_y_scale=10, rounding=25,
                                          start_x_offset=start_x_offset)
        set_graph_surface(self.latency_graph, surface)

        if features.audio and SHOW_SOUND_STATS and self.client.audio_sink:
            # audio queue graph:
            queue_values, queue_labels = norm_lists(
                (
                    (self.audio_out_queue_max, "Max"),
                    (self.audio_out_queue_cur, "Level"),
                    (self.audio_out_queue_min, "Min"),
                ),
                N_SAMPLES * 10
            )
            surface = make_graph_imagesurface(queue_values, labels=queue_labels,
                                              width=w, height=h,
                                              title="Audio Buffer (ms)", min_y_scale=10, rounding=25,
                                              start_x_offset=start_x_offset)
            set_graph_surface(self.audio_queue_graph, surface)
        return True

    def close(self, *args):
        log("SessionInfo.close(%s) is_closed=%s", args, self.is_closed)
        self.is_closed = True
        super().close()
        log("SessionInfo.close(%s) done", args)


class SessionInfoClient(InfoTimerClient):
    REFRESH_RATE = envint("XPRA_INFO_REFRESH_RATE", 2)

    def setup_connection(self, conn) -> None:
        self.session_name = self.server_session_name = "session-info"
        self.windows_enabled = False
        self.send_ping = noop
        self.server_audio_send = self.server_audio_receive = True
        self.server_audio_encoders = self.server_audio_decoders = []
        self.server_ping_latency = self.client_ping_latency = []
        self.server_start_time = 0
        super().setup_connection(conn)
        self.window = None

    def run_loop(self):
        Gtk.main()

    def exit_loop(self):
        Gtk.main_quit()

    def client_type(self):
        return "session-info"

    def update_screen(self):
        # this is called every time we get the server info back
        log("update_screen() server_last_info=%s", Ellipsizer(self.server_last_info))
        if not self.server_last_info:
            return
        td = typedict(self.server_last_info)

        def rtdict(*keys):
            d = td
            for prop in keys:
                v = d.dictget(prop)
                d = typedict(v or {})
            return d

        from xpra.client.base.serverinfo import get_remote_lib_versions
        features = rtdict("features")
        self.server_clipboard = features.boolget("clipboard")
        self.server_notifications = features.boolget("notifications")
        display = rtdict("display")
        self.server_opengl = rtdict("display", "opengl")
        self.server_bell = display.boolget("bell")
        self.server_cursors = display.boolget("cursors")
        self.server_randr = display.boolget("randr")
        server = rtdict("server")
        self.server_actual_desktop_size = server.inttupleget("root_window_size")
        self.server_max_desktop_size = server.inttupleget("max_desktop_size")
        self._remote_lib_versions = get_remote_lib_versions(rtdict("server"))
        encodings = rtdict("encodings")
        self.server_core_encodings = encodings.strtupleget("core")
        network = rtdict("network")
        self.server_packet_encoders = network.strtupleget("encoders")
        self.server_compressors = network.strtupleget("compressors")
        self.server_load = server.inttupleget("load")
        self._remote_hostname = server.strget("hostname")
        build = rtdict("server", "build")
        self._remote_version = build.strget("version")
        self._remote_build_date = build.strget("date")
        self._remote_build_time = build.strget("time")
        self._remote_revision = build.intget("revision")
        self._remote_modifications = build.intget("local_modifications")
        self._remote_branch = build.strget("branch")
        self._remote_commit = build.strget("commit")
        platform_info = rtdict("server", "platform")
        self._remote_platform = platform_info.strget("")
        self._remote_platform_release = platform_info.strget("release")
        self._remote_platform_platform = platform_info.strget("platform")
        self._remote_platform_linux_distribution = platform_info.strget("linux_distribution")
        self.server_display = server.strget("display")
        GLib.idle_add(self.do_update_screen)

    def do_update_screen(self):
        if not self.window:
            self.window = SessionInfo(self, "session-info", self._protocol._conn, show_client=False)
            self.window.show_all()

            def destroy(*_args):
                self.exit_loop()

            self.window.connect("destroy", destroy)
        self.window.populate()
        return False



#!/usr/bin/env python3

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk	#pylint: disable=wrong-import-position @UnresolvedImport


def main():
	window = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
	window.set_size_request(320, 500)
	window.connect("delete_event", Gtk.main_quit)
	vbox = Gtk.VBox(homogeneous=False, spacing=0)

	b = Gtk.Button("Modal Window")
	def show_modal_window(*_args):
		modal = Gtk.Window(type=Gtk.WindowType.TOPLEVEL)
		modal.set_transient_for(window)
		modal.set_modal(True)
		modal.show()
	b.connect('clicked', show_modal_window)
	vbox.add(b)
	window.add(vbox)
	window.show_all()
	Gtk.main()
	return 0


if __name__ == "__main__":
	main()



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Dict
from ctypes import CDLL, c_uint64, c_int, c_void_p, byref, POINTER

from xpra.common import noop
from xpra.log import Logger

from libc.stddef cimport wchar_t
from libc.stdint cimport uint64_t, uintptr_t
from libc.string cimport memset

from xpra.codecs.amf.amf cimport (
    AMF_RESULT, AMF_OK,
    AMFFactory, AMFGuid, AMFTrace, AMFSurface, AMFPlane, AMFData, AMFDataVtbl,
    AMFVariantInit,
    AMFTraceWriter, AMFTraceWriterVtbl, AMF_TRACE_DEBUG, AMF_TRACE_INFO,
    amf_uint32, amf_uint16, amf_uint8, amf_int32, amf_bool,
    AMFCaps, AMFIOCaps,
    AMF_FRAME_TYPE,
    AMF_SURFACE_FORMAT, AMF_MEMORY_TYPE,
    SURFACE_FORMAT_STR, MEMORY_TYPE_STR, DATA_TYPE_STR, FRAME_TYPE_STR,
    AMF_ACCELERATION_TYPE, ACCEL_TYPE_STR, PLANE_TYPE_STR,
    RESULT_STR,
)

cdef extern from "core/Factory.h":
    cdef const char* AMF_DLL_NAMEA


cdef extern from "Python.h":
    object PyUnicode_FromWideChar(wchar_t *w, Py_ssize_t size)
    wchar_t* PyUnicode_AsWideCharString(object unicode, Py_ssize_t *size)
    void PyMem_Free(void *ptr)


cdef extern from "string.h":
    size_t wcslen(const wchar_t *str)


log = Logger("amf")


cdef object load_library():
    libname = AMF_DLL_NAMEA.decode("latin1")
    options = [libname]
    import os
    import platform
    arch = platform.machine()   # ie: "x86_64"
    for lib_path in ("/opt/amdgpu-pro/lib64", f"/opt/amdgpu-pro/lib/{arch}-linux-gnu", os.environ.get("AMF_LIB_PATH", "")):
        if lib_path and os.path.exists(lib_path) and os.path.isdir(lib_path):
            options.append(os.path.join(lib_path, f"{libname}.so"))
    for option in options:
        try:
            return CDLL(option)
        except (OSError, ImportError) as e:
            log("CDLL({option!r})", exc_info=True)
    raise ImportError(f"AMF library {libname!r} not found")


amf = load_library()


cdef AMFFactory *factory
cdef int initialized = 0
cdef U_XPRA_TRACER = "xpra-tracer"

cdef wchar_t *XPRA_TRACER = PyUnicode_AsWideCharString(U_XPRA_TRACER, NULL)

cdef void trace_write(AMFTraceWriter* pThis, const wchar_t* scope, const wchar_t* message) noexcept nogil:
    with gil:
        scope_str = PyUnicode_FromWideChar(scope, -1) or ""
        message_str = (PyUnicode_FromWideChar(message, -1) or "").rstrip("\n\r")
        parts = message_str.split(": ", 1)
        # ie: "2025-02-25 17:46:29.642     1984 [AMFEncoderCoreH264]   Debug: AMFEncoderCoreH264Impl::Terminate()"
        # log.info(f"{scope_str=} {message_str=}")
        fn = log.info
        if len(parts) == 2:
            category = parts[0].split(" ")[-1].lower()
            message_str = parts[1]
            if (
                message_str.endswith("Switching to AllocBufferEx()")
                or message_str.startswith("Video core bandwidth calcs is not available")
                or message_str.startswith("***Found regpath, but key not found")
                or message_str.startswith("AMFEncoderCoreBaseImpl::QueryThroughput")
            ):
                fn = log.debug
            else:
                fn = {
                    "debug": log.debug,
                    "info": log.info,
                    "warning": log.warn,
                    "error": log.error,
                }.get(category, fn)
        fn(f"{scope_str}: {message_str}")

cdef void trace_flush(AMFTraceWriter* pThis) noexcept nogil:
    pass


cdef AMFTraceWriterVtbl xpra_trace_writer_functions
xpra_trace_writer_functions.Write = trace_write
xpra_trace_writer_functions.Flush = trace_flush

cdef AMFTraceWriter xpra_trace_writer
xpra_trace_writer.pVtbl = &xpra_trace_writer_functions

cdef AMFTrace *trace = NULL


cdef AMFFactory *get_factory():
    global initialized, factory
    cdef AMF_RESULT res = AMF_OK
    if not initialized:
        initialized = 1
        log("amf.common.get_factory() version=%s", get_version())
        version = get_c_version()
        AMFInit = amf.AMFInit
        AMFInit.argtypes = [c_uint64, c_void_p]
        res = AMFInit(version, <uintptr_t> &factory)
        log(f"AMFInit: {res=}")
        check(res, "AMF initialization")
        res = factory.pVtbl.GetTrace(factory, &trace)
        log(f"amf_encoder_init() GetTrace()={res}")
        if res == 0:
            assert trace != NULL
            log(f"Trace.GetGlobalLevel=%i", trace.pVtbl.GetGlobalLevel(trace))
            trace.pVtbl.RegisterWriter(trace, XPRA_TRACER, &xpra_trace_writer, 1)
            if log.is_debug_enabled():
                trace.pVtbl.SetGlobalLevel(trace, AMF_TRACE_DEBUG)
            else:
                trace.pVtbl.SetGlobalLevel(trace, AMF_TRACE_INFO)
    return factory


cdef void cleanup(self):
    global initialized
    log("amf.common.cleanup() initialized=%s", bool(initialized))
    if initialized:
        initialized = 0
        if trace:
            trace.pVtbl.UnregisterWriter(trace, XPRA_TRACER)


cdef get_version():
    version = get_c_version()
    return version >> 48, (version >> 32) & 0xffff, (version >> 16) & 0xffff, version & 0xffff


cdef void check(AMF_RESULT res, message):
    log(f"check({res}, {message!r})")
    if res == 0:
        return
    error = error_str(res) or f"error {res}"
    raise RuntimeError(f"{message}: {error}")


cdef object error_str(AMF_RESULT result):
    if result == 0:
        return ""
    # try direct code lookup:
    err = RESULT_STR(result)
    if err:
        return err
    # fallback to API call:
    cdef AMFTrace *trace = NULL
    cdef AMF_RESULT res = factory.pVtbl.GetTrace(factory, &trace)
    if res != 0:
        return ""
    cdef const wchar_t *text = trace.pVtbl.GetResultText(trace, result)
    cdef size_t size = wcslen(text)
    return PyUnicode_FromWideChar(text, size)


cdef void set_guid(AMFGuid *guid,
                   amf_uint32 _data1, amf_uint16 _data2, amf_uint16 _data3,
                   amf_uint8 _data41, amf_uint8 _data42, amf_uint8 _data43, amf_uint8 _data44,
                   amf_uint8 _data45, amf_uint8 _data46, amf_uint8 _data47, amf_uint8 _data48):
    guid.data1 = _data1
    guid.data2 = _data2
    guid.data3 = _data3
    guid.data41 = _data41
    guid.data42 = _data42
    guid.data43 = _data43
    guid.data44 = _data44
    guid.data45 = _data45
    guid.data46 = _data46
    guid.data47 = _data47
    guid.data48 = _data48


cdef uint64_t get_c_version():
    version = c_uint64()
    AMFQueryVersion = amf.AMFQueryVersion
    AMFQueryVersion.argtypes = [POINTER(c_uint64)]
    AMFQueryVersion.restype = c_int
    res = AMFQueryVersion(byref(version))
    if res:
        return 0
    return int(version.value)


cdef void fill_nv12_surface(AMFSurface *surface, amf_uint8 Y, amf_uint8 U, amf_uint8 V):
    cdef AMFPlane *planeY = surface.pVtbl.GetPlaneAt(surface, 0)
    cdef amf_int32 widthY = planeY.pVtbl.GetWidth(planeY)
    cdef amf_int32 heightY = planeY.pVtbl.GetHeight(planeY)
    cdef amf_int32 lineY = planeY.pVtbl.GetHPitch(planeY)
    cdef amf_uint8 *Ydata = <amf_uint8 *> planeY.pVtbl.GetNative(planeY)
    cdef amf_int32 y
    cdef amf_uint8 *line
    for y in range(heightY):
        line = Ydata + y * lineY
        memset(line, Y, widthY)
    cdef AMFPlane *planeUV = surface.pVtbl.GetPlaneAt(surface, 1)
    cdef amf_int32 widthUV = planeUV.pVtbl.GetWidth(planeUV)
    cdef amf_int32 heightUV = planeUV.pVtbl.GetHeight(planeUV)
    cdef amf_int32 lineUV = planeUV.pVtbl.GetHPitch(planeUV)
    cdef amf_uint8 *UVdata = <amf_uint8 *> planeUV.pVtbl.GetNative(planeUV)
    cdef amf_int32 x
    for y in range(heightUV):
        line = UVdata + y * lineUV
        for x in range(widthUV):
            line[x] = U
            line[x+1] = V


cdef object get_caps(AMFCaps *caps, props: Dict):
    cdef AMFVariantStruct var
    cdef AMF_RESULT r

    def has_property(prop: str) -> bool:
        cdef wchar_t *wprop = PyUnicode_AsWideCharString(prop, NULL)
        try:
            return caps.pVtbl.HasProperty(caps, wprop)
        finally:
            PyMem_Free(wprop)

    def query_variant(prop: str) -> None:
        check(AMFVariantInit(&var), "AMF variant initialization")
        cdef wchar_t *wprop = PyUnicode_AsWideCharString(prop, NULL)
        check(caps.pVtbl.GetProperty(caps, wprop, &var), f"query {prop} caps")
        PyMem_Free(wprop)

    def get_int64(prop: str) -> int:
        query_variant(prop)
        return int(var.int64Value)

    def get_bool(prop: str) -> bool:
        query_variant(prop)
        return bool(var.boolValue)

    pycaps = {}
    cdef AMF_ACCELERATION_TYPE accel = caps.pVtbl.GetAccelerationType(caps)
    pycaps["acceleration"] = ACCEL_TYPE_STR(accel)

    getters: Dict[str, Callable] = {
        "max-bitrate": get_int64,
        "number-of-streams": get_int64,
        "max-profile": get_int64,
        "max-tier": get_int64,
        "max-level": get_int64,
        "b-frames": get_bool,
        "fixed-sliced-mode": get_bool,
        "hardware-instances": get_int64,
        "color-conversion": get_bool,
        "max-throughput": get_int64,
        "max-bitrate": get_int64,
        "pre-analysis": get_bool,
        "requested-throughput": get_int64,
        "roi": get_bool,
        "tile-output": get_bool,
        "width-alignment": get_int64,
        "height-alignment": get_int64,
    }
    for pyname, amfname in props.items():
        if not has_property(amfname):
            continue
        getter = getters.get(pyname, noop)
        log(f"getter({pyname})={getter}")
        if getter != noop:
            try:
                value = getter(amfname)
            except RuntimeError as e:
                log(f"failed to query {pyname}: {e}")
            else:
                pycaps[pyname] = value
    # add IO caps:
    cdef AMFIOCaps *iocaps = NULL
    check(caps.pVtbl.GetInputCaps(caps, &iocaps), f"retrieving encoder input caps")
    pycaps["input"] = get_io_caps(iocaps)
    check(caps.pVtbl.GetOutputCaps(caps, &iocaps), f"retrieving encoder output caps")
    pycaps["output"] = get_io_caps(iocaps)
    return pycaps


cdef object get_io_caps(AMFIOCaps *iocaps):
    caps = {
        "vertical-align": iocaps.pVtbl.GetVertAlign(iocaps),
    }
    cdef amf_int32 minval, maxval
    iocaps.pVtbl.GetWidthRange(iocaps, &minval, &maxval)
    caps["width-range"] = (minval, maxval)
    iocaps.pVtbl.GetHeightRange(iocaps, &minval, &maxval)
    caps["height-range"] = (minval, maxval)

    cdef amf_int32 count = iocaps.pVtbl.GetNumOfFormats(iocaps)
    cdef AMF_SURFACE_FORMAT surface_format
    cdef amf_bool native
    cdef AMF_RESULT r
    formats = []
    native_formats = []
    for i in range(count):
        r = iocaps.pVtbl.GetFormatAt(iocaps, i, &surface_format, &native)
        if r:
            continue
        name = SURFACE_FORMAT_STR(surface_format)
        if native:
            native_formats.append(name)
        else:
            formats.append(name)
    caps["native-formats"] = tuple(native_formats)
    caps["formats"] = tuple(formats)
    count = iocaps.pVtbl.GetNumOfMemoryTypes(iocaps)
    cdef AMF_MEMORY_TYPE memory_type
    memory_types = []
    native_memory_types = []
    for i in range(count):
        r = iocaps.pVtbl.GetMemoryTypeAt(iocaps, i, &memory_type, &native)
        if r:
            continue
        name = MEMORY_TYPE_STR(memory_type)
        if native:
            native_memory_types.append(name)
        else:
            memory_types.append(name)
    caps["native-memory-types"] = tuple(native_memory_types)
    caps["memory-types"] = tuple(memory_types)
    caps["interlaced-supported"] = bool(iocaps.pVtbl.IsInterlacedSupported(iocaps))
    return caps


cdef object get_plane_info(AMFPlane *plane):
    assert plane
    ptype = plane.pVtbl.GetType(plane)
    cdef const AMFPlaneVtbl *pfn = plane.pVtbl
    return {
        "type": PLANE_TYPE_STR(ptype),
        "native": <uintptr_t> pfn.GetNative(plane),
        "size": pfn.GetPixelSizeInBytes(plane),
        "offset-x": pfn.GetOffsetX(plane),
        "offset-y": pfn.GetOffsetY(plane),
        "width": pfn.GetWidth(plane),
        "height": pfn.GetHeight(plane),
        "h-pitch": pfn.GetHPitch(plane),
        "v-pitch": pfn.GetVPitch(plane),
        "is-tiled": pfn.IsTiled(plane),
    }


cdef object get_data_info(AMFData *data):
    assert data
    cdef const AMFDataVtbl *dfn = data.pVtbl
    return {
        "property-count": dfn.GetPropertyCount(data),
        "memory-type": MEMORY_TYPE_STR(dfn.GetMemoryType(data)),
        "data-type": DATA_TYPE_STR(dfn.GetDataType(data)),
    }


cdef object get_surface_info(AMFSurface *surface):
    assert surface
    cdef const AMFSurfaceVtbl *sfn = surface.pVtbl
    fmt = sfn.GetFormat(surface)
    cdef AMF_FRAME_TYPE ftype = sfn.GetFrameType(surface)
    return {
        "format": SURFACE_FORMAT_STR(fmt),
        "planes": sfn.GetPlanesCount(surface),
        "frame-type": FRAME_TYPE_STR(ftype),
    }



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable, Sequence

from xpra.platform import platform_import

SHADOW_OPTIONS: dict[str, Callable[[], bool]] = {}

GSTREAMER_CAPTURE_ELEMENTS: Sequence[str] = ()


def ShadowServer(*_args):  # pragma: no cover
    raise NotImplementedError()


platform_import(globals(), "shadow_server", True,
                "ShadowServer",
                "SHADOW_OPTIONS",
                "GSTREAMER_CAPTURE_ELEMENTS",
                )



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Callable, Sequence

from xpra.keyboard.common import KeyEvent
from xpra.client.gui.keyboard_shortcuts_parser import parse_shortcut_modifiers, parse_shortcuts, get_modifier_names
from xpra.util.str_fn import std, csv, Ellipsizer
from xpra.util.env import envbool
from xpra.common import noop
from xpra.log import Logger

log = Logger("keyboard")

LAYOUT_GROUPS = envbool("XPRA_LAYOUT_GROUPS", True)
DEBUG_KEY_EVENTS = tuple(x.strip().lower() for x in os.environ.get("XPRA_DEBUG_KEY_EVENTS", "").split(",") if x.strip())


class KeyboardHelper:

    def __init__(self, net_send: Callable, keyboard_sync=True,
                 shortcut_modifiers="auto", key_shortcuts=(),
                 raw=False, backend="", model="", layout="", layouts=(),
                 variant="", variants=(), options=""):
        self.reset_state()
        self.send = net_send
        self.locked = False
        self.sync = keyboard_sync
        self.shortcuts_enabled = True
        self.shortcut_modifiers_str = shortcut_modifiers
        self.shortcut_modifiers: list[str] = []
        self.key_shortcuts_strs = key_shortcuts
        self.key_shortcuts: dict[str, list[str]] = {}
        # command line overrides:
        self.raw = raw
        self.backend = backend
        self.model_option = model
        self.layout_option = layout if (layout or "").lower() not in ("client", "auto") else ""
        self.variant_option = variant
        self.layouts_option = layouts
        self.variants_option = variants
        self.options = options
        # the platform class which allows us to map the keys:
        from xpra.platform.keyboard import Keyboard  # pylint: disable=import-outside-toplevel
        self.keyboard = Keyboard()  # pylint: disable=not-callable
        log("KeyboardHelper(%s) keyboard=%s",
            (net_send, keyboard_sync, key_shortcuts,
             raw, model, layout, layouts, variant, variants, options), self.keyboard)
        key_repeat = self.keyboard.get_keyboard_repeat()
        if key_repeat:
            self.key_repeat_delay, self.key_repeat_interval = key_repeat

    def set_platform_layout(self, layout: str) -> None:
        if hasattr(self.keyboard, "set_platform_layout"):
            self.keyboard.set_platform_layout(layout)

    def mask_to_names(self, mask) -> list[str]:
        return self.keyboard.mask_to_names(mask)

    def set_modifier_mappings(self, mappings: dict[str, str]):
        self.keyboard.set_modifier_mappings(mappings)

    def reset_state(self) -> None:
        self.keycodes: Sequence[tuple[int, str, int, int, int]] = ()
        self.x11_keycodes: dict[int, list[str]] = {}
        self.mod_meanings: dict[str, Any] = {}
        self.mod_managed: list[str] = []
        self.mod_pointermissing: list[str] = []
        self.backend = ""
        self.model = ""
        self.layout = ""
        self.layouts: list[str] = []
        self.variant = ""
        self.variants: list[str] = []
        self.options = ""
        self.backend = ""
        self.name = ""
        self.query = ""
        self.query_struct: dict[str, Any] = {}
        self.layout_groups = LAYOUT_GROUPS
        self.raw = False

        self.hash = None

        self.key_repeat_delay = -1
        self.key_repeat_interval = -1
        self.sync = False
        self.key_shortcuts = {}

    def cleanup(self) -> None:
        self.reset_state()
        self.send = noop

    def keymap_changed(self, *args) -> None:
        """ This method is overridden in the GTK Keyboard Helper """

    def parse_shortcuts(self) -> dict[str, list]:
        modifier_names = self.get_modifier_names()
        self.shortcut_modifiers = parse_shortcut_modifiers(self.shortcut_modifiers_str, modifier_names)
        self.key_shortcuts = parse_shortcuts(self.key_shortcuts_strs, self.shortcut_modifiers, modifier_names)
        return self.key_shortcuts

    def get_modifier_names(self) -> dict[str, str]:
        return get_modifier_names(self.mod_meanings)

    def key_handled_as_shortcut(self, window, key_name: str, modifiers: list[str], depressed: bool) -> bool:
        # find the shortcuts that may match this key:
        shortcuts = self.key_shortcuts.get(key_name)
        log("key_handled_as_shortcut%s shortcuts_enabled=%s, shortcuts=%s",
            (window, key_name, modifiers, depressed),
            self.shortcuts_enabled, shortcuts)
        if not self.shortcuts_enabled:
            return False
        if not shortcuts:
            return False
        if len(shortcuts) > 1:
            # sort shortcuts based on how many modifiers are required,
            # so that if multiple shortcuts use the same key,
            # we will try to match the one with the most modifiers first.
            # ie: Num_Lock+Menu will be tested before Menu
            # (this is needed because Num_Lock is then discarded when comparing the list of required modifiers!)
            shortcuts = sorted(shortcuts, key=lambda x: len(x[0]), reverse=True)
        for shortcut in shortcuts:
            if self._check_shortcut(window, key_name, modifiers, depressed, shortcut):
                return True
        return False

    def _check_shortcut(self, window, key_name: str, modifiers: list[str], depressed: bool, shortcut: Sequence) -> bool:
        req_mods, action, args = shortcut
        extra_modifiers = list(modifiers)
        for rm in req_mods:
            if rm not in modifiers:
                # modifier is missing, bail out
                log("not matched %s for %s: %s not in %s",
                    shortcut, key_name, rm, modifiers)
                return False
            try:
                extra_modifiers.remove(rm)
            except ValueError:
                pass  # same modifier listed twice?
        kmod, _, ignored = self.keyboard.get_keymap_modifiers()
        if not kmod and self.keyboard.modifier_keys:
            # fallback to server supplied map:
            kmod = self.keyboard.modifier_keys
        # ie: {'ISO_Level3_Shift': 'mod5', 'Meta_L': 'mod1', ...}
        log("keymap modifiers: %s", kmod)
        ignoremod = ("Caps_Lock", "Num_Lock")
        for x in ignoremod:
            mod = kmod.get(x)
            if mod in extra_modifiers:
                extra_modifiers.remove(mod)
        for mod in ignored:
            if mod in extra_modifiers:
                extra_modifiers.remove(mod)
        if extra_modifiers:
            log("skipping partial shortcut match %s, modifiers unmatched: %s", shortcut, extra_modifiers)
            return False
        log("matched shortcut %s", shortcut)
        if not depressed:
            # when the key is released, just ignore it - do NOT send it to the server!
            return True
        if action == "pass":
            return False
        try:
            method = getattr(window, action)
            log("key_handled_as_shortcut(%s,%s,%s,%s) found shortcut=%s, will call %s%s",
                window, key_name, modifiers, depressed, shortcut, method, args)
        except AttributeError as e:
            log.error("key dropped, invalid method name in shortcut %s: %s", action, e)
            return True
        try:
            method(*args)
            log("key_handled_as_shortcut(%s,%s,%s,%s) has been handled: %s",
                window, key_name, modifiers, depressed, method)
        except Exception:
            log.error("Error: key_handled_as_shortcut(%s,%s,%s,%s)", window, key_name, modifiers, depressed)
            log.error(" failed to execute shortcut=%s", shortcut)
            log.error("", exc_info=True)
        return True

    def process_key_event(self, wid: int, key_event: KeyEvent) -> bool:
        """
            This method gives the Keyboard class
            a chance to fire more than one send_key_action.
            (win32 uses this for AltGr emulation)
        """
        self.keyboard.process_key_event(self.send_key_action, wid, key_event)
        return False

    def debug_key_event(self, wid: int, key_event: KeyEvent) -> None:
        if not DEBUG_KEY_EVENTS:
            return

        def keyname(v: str) -> str:
            if v.endswith("_L") or v.endswith("_R"):
                return v[:-2].lower()
            return v.lower()

        def dbg(v) -> bool:
            return v and keyname(v) in DEBUG_KEY_EVENTS

        debug = ("all" in DEBUG_KEY_EVENTS) or dbg(key_event.keyname) or dbg(key_event.string)
        modifiers = key_event.modifiers
        if not debug and modifiers:
            # see if one of the modifier matches:
            # either the raw name (ie: "mod2") or its actual meaning (ie: "NumLock")
            for m in modifiers:
                if m in DEBUG_KEY_EVENTS:
                    debug = True
                    break
                name = keyname(self.keyboard.modifier_names.get(m) or "")
                if name and name in DEBUG_KEY_EVENTS:
                    debug = True
                    break
        if debug:
            log.info("key event %s on window %i", key_event, wid)

    def send_key_action(self, wid: int, key_event: KeyEvent) -> None:
        log("send_key_action(%s, %s)", wid, key_event)
        packet = ["key-action", wid]
        for x in ("keyname", "pressed", "modifiers", "keyval", "string", "keycode", "group"):
            packet.append(getattr(key_event, x))
        self.debug_key_event(wid, key_event)
        self.send(*packet)

    def get_layout_spec(self) -> tuple[str, str, Sequence[str], str, Sequence[str], str]:
        """ add / honour overrides """
        model, layout, layouts, variant, variants, options = self.keyboard.get_layout_spec()
        log("%s.get_layout_spec()=%s", self.keyboard, (layout, layouts, variant, variants, options))

        def inl(v, l) -> list:
            try:
                if v in l or v is None:
                    return l
                return [v] + list(l)
            except Exception:
                if v is not None:
                    return [v]
                return []

        layout = self.layout_option or layout
        layouts = inl(layout, self.layouts_option or layouts)
        variant = self.variant_option or variant
        variants = inl(variant, self.variants_option or variants)
        val = (
            self.model_option or model,
            layout,
            layouts,
            variant,
            variants,
            self.options or options,
        )
        log("get_layout_spec()=%s", val)
        return val

    def get_keymap_spec(self) -> dict[str, Any]:
        query_struct = self.keyboard.get_keymap_spec()
        if query_struct:
            if self.backend:
                query_struct["backend"] = self.backend
            if self.model_option:
                query_struct["model"] = self.model_option
            if self.layout_option:
                query_struct["layout"] = self.layout_option
            if self.layouts_option:
                query_struct["layouts"] = csv(self.layouts_option)
            if self.variant_option:
                query_struct["variant"] = self.variant_option
            if self.variants_option:
                query_struct["variants"] = csv(self.variants_option)
            if self.options:
                if self.options.lower() == "none":
                    query_struct["options"] = ""
                else:
                    query_struct["options"] = self.options
        return query_struct

    def query_xkbmap(self) -> None:
        log("query_xkbmap()")
        self.model, self.layout, self.layouts, self.variant, self.variants, self.options = self.get_layout_spec()
        self.query_struct = self.get_keymap_spec()
        log(f"query_xkbmap() query_struct={self.query_struct}")
        self.keycodes = self.get_full_keymap()
        log(f"query_xkbmap() keycodes={self.keycodes}")
        self.x11_keycodes = self.keyboard.get_x11_keymap()
        log(f"query_xkbmap() {self.keyboard}.get_x11_keymap()={self.x11_keycodes}")
        mods = self.keyboard.get_keymap_modifiers()
        self.mod_meanings, self.mod_managed, self.mod_pointermissing = mods
        log(f"query_xkbmap() get_keymap_modifiers()={mods}")
        self.update_hash()
        log(f"layout={self.layout}, layouts={self.layouts}, variant={self.variant}, variants={self.variants}")
        log(f"query-struct={self.query_struct}")
        log("keycodes=%s", Ellipsizer(self.keycodes))
        log("x11 keycodes=%s", Ellipsizer(self.x11_keycodes))
        log(f"mod managed: {self.mod_managed}")
        log(f"mod meanings: {self.mod_meanings}")
        log(f"mod pointermissing: {self.mod_pointermissing}")
        log(f"hash={self.hash}")

    def update(self) -> None:
        if not self.locked:
            self.query_xkbmap()
            self.parse_shortcuts()

    def layout_str(self) -> str:
        if self.backend and self.name:
            return f"{self.backend} {self.name!r}"
        return " / ".join(str(x) for x in (
            self.layout_option or self.layout, self.variant_option or self.variant) if bool(x))

    def send_layout(self) -> None:
        log("send_layout() layout_option=%r, layout=%r, variant_option=%r, variant=%r, options=%r",
            self.layout_option, self.layout, self.variant_option, self.variant, self.options)
        log(f"send_layout() backend={self.backend!r}, name={self.name!r}")
        self.send("layout-changed",
                  self.layout_option or self.layout or "",
                  self.variant_option or self.variant or "",
                  self.options or "",
                  self.backend or "", self.name or "")

    def send_keymap(self) -> None:
        log("send_keymap()")
        props = {"keymap": self.get_keymap_properties()}
        self.send("keymap-changed", props)

    def update_hash(self) -> None:
        import hashlib
        h = hashlib.sha256()

        def hashadd(v) -> None:
            h.update(("/%s" % str(v)).encode("utf8"))

        for x in (self.mod_meanings, self.mod_pointermissing, self.keycodes, self.x11_keycodes):
            hashadd(x)
        if self.query_struct:
            # flatten the dict in a predicatable order:
            for k in sorted(self.query_struct.keys()):
                hashadd(self.query_struct.get(k))
        self.hash = "/".join([str(x) for x in (self.layout, self.variant, h.hexdigest()) if bool(x)])

    def get_full_keymap(self) -> Sequence[tuple[int, str, int, int, int]]:
        return ()

    def get_keymap_properties(self, skip=()) -> dict[str, Any]:
        props = {}
        for x in (
                "backend", "layout", "layouts", "variant", "variants",
                "raw", "layout_groups",
                "sync",
                "query_struct", "mod_meanings",
                "mod_managed", "mod_pointermissing", "keycodes", "x11_keycodes",
        ):
            if x in skip:
                continue
            v = getattr(self, x)
            if v:
                props[x] = v
        return props

    def log_keyboard_info(self) -> None:
        # show the user a summary of the settings used
        # (values detected with overrides applied)
        model, layout, _, variant, _, options = self.get_layout_spec()
        kb_info = {
            "model": model,
            "layout": layout,
            "variant": variant,
            "options": options,
        }
        if self.backend:
            kb_info["backend"] = self.backend
        if self.query_struct:
            kb_info["rules"] = self.query_struct.get("rules", "")
        if not kb_info:
            log.info(" using default keyboard settings")
        else:
            log.info(" keyboard settings: %s", csv(f"{k}={std(v)}" for k, v in kb_info.items() if v))



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct
import cairo
import unittest
from time import monotonic
from io import BytesIO
from PIL import Image

from xpra.codecs.checks import TEST_COLORS
from xpra.gtk.cairo_image import make_image_surface


N = 5
WIDTH, HEIGHT = 256, 256


def surface_to_pil(img) -> Image:
    save = BytesIO()
    img.write_to_png(save)
    pixels = save.getvalue()
    save.close()
    return Image.open(BytesIO(pixels))


class CairoColorsTest(unittest.TestCase):

    def test_compare(self):
        for color_name, hex_pixel in TEST_COLORS.items():
            # color components as integers:
            r, g, b, a = (int(hex_pixel[i * 2:(i + 1) * 2], 16) for i in range(4))
            # color components as bytes:
            rb, gb, bb, ab = (struct.pack("B", pixel_int) for pixel_int in (r, g, b, a))
            # map component to byte values:
            pix_vals: dict[str, bytes] = {
                "R": rb,
                "G": gb,
                "B": bb,
                "A": ab,
                "X": ab,
            }
            for format_name, cairo_format in {
                "ARGB32": cairo.Format.ARGB32,
                "RGB24": cairo.Format.RGB24,
            }.items():
                # try with our code:
                for rgb_format in ("RGBA", "RGBX", "BGRA", "BGRX", "RGB", "BGR"):
                    stride = WIDTH * len(rgb_format)
                    pixels = (b"".join(pix_vals[component] for component in rgb_format)) * stride * HEIGHT
                    start = monotonic()
                    for _ in range(N):
                        img = make_image_surface(cairo_format, rgb_format, pixels, WIDTH, HEIGHT, stride)
                    end = monotonic()
                    if color_name == "blue":
                        mps = round(N*WIDTH*HEIGHT / (end-start) / 1024 / 1024)
                        print(f"{rgb_format:<4} to cairo.{format_name: <6} : {mps:>5} MPixels/s")
                    pimg = surface_to_pil(img)
                    pixel = pimg.getpixel((WIDTH//2, HEIGHT//2))
                    if pixel != (r, g, b):
                        raise ValueError(f"expected {(r, g, b)} but got {pixel}")

                # try painting the same color using cairo
                img = cairo.ImageSurface(cairo_format, WIDTH, HEIGHT)
                ctx = cairo.Context(img)
                ctx.set_source_rgb(r/255, g/255, b/255)
                ctx.rectangle(0, 0, WIDTH, HEIGHT)
                ctx.fill()
                # img.write_to_png(f"{color_name}-cairo.png")
                pimg = surface_to_pil(img)
                pixel = pimg.getpixel((WIDTH//2, HEIGHT//2))
                if pixel != (r, g, b):
                    raise ValueError(f"expected {(r, g, b)} but got {pixel}")


def main():
    unittest.main()


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2023 Chris Marchetti <adamnew123456@gmail.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Optional
from collections.abc import Callable

get_parent_pid: Optional[Callable] = None
try:
    from xpra.platform.posix import proc_libproc

    get_parent_pid = proc_libproc.get_parent_pid
except (ImportError, AttributeError):
    try:
        from xpra.platform.posix import proc_procps

        get_parent_pid = proc_procps.get_parent_pid
    except (ImportError, AttributeError):
        pass


def main(argv) -> int:
    from xpra.platform import program_context
    with program_context("Get-Parent-Pid", "Get Parent Pid"):
        if not callable(get_parent_pid):
            print("`get_parent_pid` is not available!")
            return 1
        print(f"using `get_parent_pid`={get_parent_pid}")
        try:
            print(f"from {get_parent_pid.__module__}")
        except AttributeError:  # `__module__` is CPython only?
            pass
        for pid_str in argv[1:]:
            try:
                pid = int(pid_str)
            except ValueError:
                print(f"{pid_str} is not a valid pid number")
            else:
                print(f" get_parent_pid({pid})={get_parent_pid(pid)}")
    return 0


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
import glob
import socket
import errno
from collections.abc import Iterator

from xpra.common import SocketState
from xpra.util.env import osexpand
from xpra.util.io import is_socket, umask_context, get_util_logger
from xpra.os_util import POSIX
from xpra.platform.dotxpra_common import PREFIX
from xpra.platform import platform_import

DISPLAY_PREFIX = ":"


def norm_makepath(dirpath: str, name: str) -> str:
    return os.path.join(dirpath, PREFIX + strip_display_prefix(name))


def strip_display_prefix(s: str) -> str:
    if s.startswith(DISPLAY_PREFIX):
        return s[len(DISPLAY_PREFIX):]
    return s


def debug(msg: str, *args, **kwargs) -> None:
    log = get_util_logger()
    log(msg, *args, **kwargs)


class DotXpra:
    def __init__(self, sockdir="", sockdirs=None, actual_username="", uid=0, gid=0):
        self.uid = uid or os.getuid()
        self.gid = gid or os.getgid()
        self.username = actual_username
        sockdirs = sockdirs or []
        if not sockdir:
            if sockdirs:
                sockdir = sockdirs[0]
            else:
                sockdir = "undefined"
        elif sockdir not in sockdirs:
            sockdirs.insert(0, sockdir)
        self._sockdir = self.osexpand(sockdir)
        self._sockdirs: list[str] = [self.osexpand(x) for x in sockdirs]

    def osexpand(self, v: str) -> str:
        return osexpand(v, self.username, self.uid, self.gid)

    def __repr__(self):
        return f"DotXpra({self._sockdir}, {self._sockdirs} - {self.uid}:{self.gid} - {self.username})"

    def mksockdir(self, d: str, mode=0o700, uid=None, gid=None) -> None:
        if not d:
            return
        if not os.path.exists(d):
            if uid is None:
                uid = self.uid
            if gid is None:
                gid = self.gid
            parent = os.path.dirname(d)
            if parent and parent != "/" and not os.path.exists(parent):
                self.mksockdir(parent, mode, uid, gid)
            with umask_context(0):
                os.mkdir(d, mode)
            if uid != os.getuid() or gid != os.getgid():
                os.lchown(d, uid, gid)
        elif d != "/tmp":
            try:
                st_mode = os.stat(d).st_mode & 0o777
                if st_mode == 0o750 and d.startswith("/run/xpra"):
                    # this directory is for shared sockets, o750 is OK
                    return
                if st_mode != mode:
                    # perhaps this directory lives in $XDG_RUNTIME_DIR
                    # ie: /run/user/$UID/xpra or /run/user/$UID/xpra/100
                    xrd = os.environ.get("XDG_RUNTIME_DIR")
                    if xrd and d.startswith(xrd) and os.stat(xrd).st_mode & 0o777 == 0o700:
                        # $XDG_RUNTIME_DIR has the correct permissions
                        return
                    log = get_util_logger()
                    log.warn(f"Warning: socket directory {d!r}")
                    log.warn(f" expected permissions {mode:o} but found {st_mode:o}")
            except OSError:
                get_util_logger().error("Error: mksockdir%s", (d, mode, uid, gid), exc_info=True)

    def socket_expand(self, path: str) -> str:
        return osexpand(path, self.username, uid=self.uid, gid=self.gid)

    def norm_socket_paths(self, local_display_name: str) -> list[str]:
        return [norm_makepath(x, local_display_name) for x in self._sockdirs]

    def socket_path(self, local_display_name: str) -> str:
        return norm_makepath(self._sockdir, local_display_name)

    def get_server_state(self, sockpath: str, timeout=5) -> SocketState:
        saved_sockpath = sockpath
        if sockpath.startswith("@"):
            assert POSIX
            sockpath = "\0" + sockpath[1:]
        elif not os.path.exists(sockpath):
            return SocketState.DEAD
        sock = socket.socket(socket.AF_UNIX)
        sock.settimeout(timeout)
        try:
            sock.connect(sockpath)
            return SocketState.LIVE
        except OSError as e:
            debug(f"get_server_state: connect({saved_sockpath!r})={e} (timeout={timeout}")
            err = e.args[0]
            if err == errno.EACCES:
                return SocketState.INACCESSIBLE
            if err == errno.ECONNREFUSED:
                if saved_sockpath.startswith("@"):
                    return SocketState.DEAD
                # could be the server is starting up
                debug("ECONNREFUSED")
                return SocketState.UNKNOWN
            if err == errno.EWOULDBLOCK:
                debug("EWOULDBLOCK")
                return SocketState.DEAD
            if err == errno.ENOENT:
                debug("ENOENT")
                return SocketState.DEAD
            return SocketState.UNKNOWN
        finally:
            try:
                sock.close()
            except OSError:
                debug("%s.close()", sock, exc_info=True)

    def displays(self, check_uid=-1, matching_state=None) -> list[str]:
        return list(set(v[1] for v in self.sockets(check_uid, matching_state)))

    def sockets(self, check_uid=-1, matching_state=None) -> list:
        # flatten the dictionary into a list:
        return list(set((v[0], v[1]) for details_values in
                        self.socket_details(check_uid, matching_state).values() for v in details_values))

    def socket_paths(self, check_uid=-1, matching_state=None, matching_display=None) -> list[str]:
        paths = []
        for details in self.socket_details(check_uid, matching_state, matching_display).values():
            for _, _, socket_path in details:
                paths.append(socket_path)
        return paths

    def _unique_sock_dirs(self) -> Iterator[str]:
        dirs = []
        if self._sockdir != "undefined":
            dirs.append(self._sockdir)
        dirs += [x for x in self._sockdirs if x not in dirs]
        seen = set()
        for d in dirs:
            if d in seen or not d:
                continue
            seen.add(d)
            real_dir = os.path.realpath(osexpand(d))
            if real_dir != d:
                if real_dir in seen:
                    continue
                seen.add(real_dir)
            if not os.path.exists(real_dir):
                debug(f"socket_details: directory {real_dir!r} does not exist")
                continue
            yield real_dir

    def get_display_state(self, display: str) -> SocketState:
        state = None
        for d in self._unique_sock_dirs():
            # look for a 'socket' in the session directory
            # ie: /run/user/1000/xpra/10
            session_dir = os.path.join(d, strip_display_prefix(display))
            if os.path.exists(session_dir):
                # ie: /run/user/1000/xpra/10/socket
                sockpath = os.path.join(session_dir, "socket")
                if os.path.exists(sockpath):
                    state = self.is_socket_match(sockpath)
                    if state is SocketState.LIVE:
                        return state
            # when not using a session directory,
            # add the prefix to prevent clashes on NFS:
            # ie: "~/.xpra/HOSTNAME-10"
            sockpath = os.path.join(d, PREFIX + strip_display_prefix(display))
            state = self.is_socket_match(sockpath)
            if state is SocketState.LIVE:
                return state
        return state or SocketState.DEAD

    # find the matching sockets, and return:
    # (state, local_display, sockpath) for each socket directory we probe
    def socket_details(self, check_uid=-1, matching_state=None, matching_display=None) \
            -> dict[str, list[tuple[SocketState, str, str]]]:
        sd: dict[str, list[tuple[SocketState, str, str]]] = {}
        debug("socket_details%s sockdir=%s, sockdirs=%s",
              (check_uid, matching_state, matching_display), self._sockdir, self._sockdirs)

        def add_result(d: str, item: tuple[SocketState, str, str]) -> None:
            results: list[tuple[SocketState, str, str]] = sd.setdefault(d, [])
            if item not in results:
                results.append(item)

        def local(display: str) -> str:
            if display.startswith("wayland-"):
                return display
            return DISPLAY_PREFIX + strip_display_prefix(display)

        def add_session_dir(session_dir: str, display: str) -> None:
            if not os.path.exists(session_dir):
                debug("add_session_dir%s path does not exist", (session_dir, display))
                return
            if not os.path.isdir(session_dir):
                debug("add_session_dir%s not a directory", (session_dir, display))
                return
            # ie: /run/user/1000/xpra/10/socket
            sockpath = os.path.join(session_dir, "socket")
            if os.path.exists(sockpath):
                state = self.is_socket_match(sockpath, matching_state=matching_state)
                debug("add_session_dir(%s) state(%s)=%s", (session_dir, display), sockpath, state)
                if state:
                    add_result(session_dir, (state, local(display), sockpath))

        for d in self._unique_sock_dirs():
            # if we know the display name,
            # we know the corresponding session dir:
            if matching_display:
                session_dir = os.path.join(d, strip_display_prefix(matching_display))
                add_session_dir(session_dir, matching_display)
            else:
                # find all the directories that could be session directories:
                for p in os.listdir(d):
                    if p.startswith("wayland-"):
                        p = p[len("wayland-"):]
                    try:
                        int(p)
                    except ValueError:
                        continue
                    session_dir = os.path.join(d, p)
                    add_session_dir(session_dir, p)
            # ie: "~/.xpra/HOSTNAME-"
            base = os.path.join(d, PREFIX)
            if matching_display:
                dstr = strip_display_prefix(matching_display)
            else:
                dstr = "*"
            potential_sockets = glob.glob(base + dstr)
            for sockpath in sorted(potential_sockets):
                state = self.is_socket_match(sockpath, check_uid, matching_state)
                if state:
                    display = local(sockpath[len(base):])
                    add_result(d, (state, display, sockpath))
        return sd

    def is_socket_match(self, sockpath: str, check_uid=-1, matching_state=None) -> SocketState | None:
        if not is_socket(sockpath, check_uid):
            return None
        state = self.get_server_state(sockpath)
        if matching_state and state != matching_state:
            debug("is_socket_match%s state '%s' does not match", (sockpath, check_uid, matching_state), state)
            return None
        return state


# win32 re-defines DotXpra for namedpipes:
platform_import(globals(), "dotxpra", False,
                "DotXpra",
                "DISPLAY_PREFIX",
                "norm_makepath")



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from math import ceil, floor

from xpra.os_util import gi_import
from xpra.log import Logger
log = Logger("bindings", "gtk")

GObject = gi_import("GObject")
log(f"GObject={GObject}")

cdef extern from "glib-2.0/glib-object.h":
    ctypedef struct cGObject "GObject":
        pass

cdef extern from "pygobject-3.0/pygobject.h":
    cGObject *pygobject_get(object box)
    object pygobject_new(cGObject * contents)

    ctypedef void* gpointer
    ctypedef int GType
    ctypedef struct PyGBoxed:
        #PyObject_HEAD
        gpointer boxed
        GType gtype


cdef object wrap(cGObject * contents):
    # Put a raw GObject* into a PyGObject wrapper.
    return pygobject_new(contents)


cdef cGObject *unwrap(box, pyclass) except? NULL:
    # Extract a raw GObject* from a PyGObject wrapper.
    assert issubclass(pyclass, GObject.GObject)
    if not isinstance(box, pyclass):
        raise TypeError("object %r is not a %r" % (box, pyclass))
    return pygobject_get(box)



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os.path
import subprocess

from xpra.os_util import POSIX, OSX, gi_import
from xpra.util.io import which
from xpra.gtk.signals import register_os_signals
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import imagebutton, label, setfont
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.util.thread import start_thread
from xpra.log import Logger

Gtk = gi_import("Gtk")
GLib = gi_import("GLib")

log = Logger("client", "util")


def exec_command(cmd):
    env = os.environ.copy()
    proc = subprocess.Popen(cmd, env=env)
    return proc


def xal(widget, xalign=1):
    al = Gtk.Alignment(xalign=xalign, yalign=0.5, xscale=0, yscale=0)
    al.add(widget)
    return al


def sf(w, font="sans 14"):
    setfont(w, font)
    return w


def l(text):  # noqa: E743
    widget = label(text)
    return sf(widget)


class DesktopGreeter(Gtk.Window):

    def __init__(self):
        self.exit_code = None
        super().__init__()
        self.set_border_width(20)
        self.set_title("Start Desktop Environment")
        self.set_position(Gtk.WindowPosition.CENTER)
        self.set_size_request(640, 300)
        icon = get_icon_pixbuf("xpra.png")
        if icon:
            self.set_icon(icon)
        self.connect("delete-event", self.quit)
        add_close_accel(self, self.quit)

        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        vbox.set_spacing(10)

        self.desktop_combo = sf(Gtk.ComboBoxText())
        vbox.add(self.desktop_combo)

        # Action buttons:
        hbox = Gtk.HBox(homogeneous=False, spacing=20)
        vbox.pack_start(hbox, False, True, 20)

        def btn(text, tooltip, callback, default=False):
            ib = imagebutton(text, tooltip=tooltip, clicked_callback=callback, icon_size=32,
                             default=default, label_font="sans 16")
            hbox.pack_start(ib)
            return ib

        self.cancel_btn = btn("Exit", "", self.quit)
        self.run_btn = btn("Start", "Start the desktop environment", self.run_command)

        vbox.show_all()
        self.add(vbox)
        start_thread(self.load_desktop_session, "load-desktop-sessions", daemon=True)

    def app_signal(self, signum):
        if self.exit_code is None:
            self.exit_code = 128 + signum
        log("app_signal(%s) exit_code=%i", signum, self.exit_code)
        self.quit()

    def quit(self, *args):
        log("quit%s", args)
        if self.exit_code is None:
            self.exit_code = 0
        self.do_quit()

    def do_quit(self):
        log("do_quit()")
        Gtk.main_quit()

    def run_command(self, *_args):
        name = self.desktop_combo.get_active_text()
        cmd = name
        if cmd in self.desktop_sessions:
            session = self.desktop_sessions.get(cmd)
            # ie:
            # session={
            #    'Type': '',
            #    'VersionString': '',
            #    'Name': 'Deepin',
            #    'GenericName': '',
            #    'NoDisplay': False,
            #    'Comment': 'Deepin Desktop Environment',
            #    'Icon': '',
            #    'Hidden': False,
            #    'OnlyShowIn': [],
            #    'NotShowIn': [],
            #    'Exec': '/usr/bin/startdde',
            #    'TryExec': '/usr/bin/startdde',
            #    'Path': '',
            #    'Terminal': False,
            #    'MimeTypes': [],
            #    'Categories': [],
            #    'StartupNotify': False,
            #    'StartupWMClass': '',
            #    'URL': '',
            #    'command': '/usr/bin/startdde',
            #    'IconData': b'<svg>...</svg>\n',
            #    'IconType': 'svg'
            #    }
            for k in ("command", "Exec", "TryExec"):
                cmd = session.get(k)
                if cmd:
                    break
        if cmd and not os.path.isabs(cmd):
            cmd = which(cmd)
        if not cmd:
            log.warn("no command found for '%s'", name)
            return
        argv = [cmd]
        self.close()
        os.execv(cmd, argv)

    def wait_for_subprocess(self, proc):
        proc.wait()
        log("return code: %s", proc.returncode)
        GLib.idle_add(self.show)

    def load_desktop_session(self):
        from xpra.server.menu_provider import get_menu_provider
        self.desktop_sessions = get_menu_provider().get_desktop_sessions()
        GLib.idle_add(self.populate_xsessions)

    def populate_xsessions(self):
        log("populate_xsessions()")
        self.desktop_combo.get_model().clear()
        if self.desktop_sessions:
            for name in sorted(self.desktop_sessions.keys()):
                self.desktop_combo.append_text(name)
        else:
            self.desktop_combo.append_text("xterm")
        self.desktop_combo.set_active(0)


def main(_options=None):  # pragma: no cover
    # pylint: disable=import-outside-toplevel
    assert POSIX and not OSX
    from xpra.platform import program_context
    from xpra.log import enable_color
    from xpra.platform.gui import init, ready
    with program_context("xpra-start-gui", "Xpra Start GUI"):
        enable_color()
        init()
        gui = DesktopGreeter()
        register_os_signals(gui.app_signal)
        ready()
        gui.show()
        gui.present()
        Gtk.main()
        log("do_main() gui.exit_code=%s", gui.exit_code)
        return gui.exit_code


if __name__ == "__main__":  # pragma: no cover
    r = main()
    sys.exit(r)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.util.objects import typedict
from typing import Any
from collections.abc import Callable

from xpra.common import NotificationID
from xpra.server.source.stub_source_mixin import StubSourceMixin
from xpra.log import Logger

log = Logger("notify")


class NotificationMixin(StubSourceMixin):

    PREFIX = "notification"

    @classmethod
    def is_needed(cls, caps: typedict) -> bool:
        v = caps.get("notifications")
        if isinstance(v, bool):
            return v
        if isinstance(v, dict):
            return typedict(v).boolget("enabled", False)
        return False

    def init_state(self) -> None:
        self.send_notifications: bool = False
        self.notification_callbacks: dict[int, Callable] = {}

    def parse_client_caps(self, c: typedict) -> None:
        v = c.get("notifications")
        if isinstance(v, dict):
            self.send_notifications = typedict(v).boolget("enabled")
        log("send notifications=%s", self.send_notifications)

    def get_info(self) -> dict[str, Any]:
        return {
            NotificationMixin.PREFIX: self.send_notifications,
        }

    ######################################################################
    # notifications:
    # Utility functions for mixins (makes notifications optional)
    def may_notify(self, nid: int | NotificationID = 0, summary: str = "", body: str = "",
                   actions=(), hints=None, expire_timeout=10 * 1000,
                   icon_name: str = "", user_callback: Callable | None = None) -> None:
        try:
            from xpra.platform.paths import get_icon_filename
            from xpra.notifications.common import parse_image_path
        except ImportError as e:
            log("not sending notification: %s", e)
        else:
            icon_filename = get_icon_filename(icon_name)
            icon = parse_image_path(icon_filename)
            self.notify("", nid, "Xpra", 0, "",
                        summary, body, actions, hints or {},
                        expire_timeout, icon, user_callback)

    def notify(self, dbus_id, nid: int, app_name: str, replaces_nid: int, app_icon,
               summary: str, body: str, actions, hints, expire_timeout: int,
               icon, user_callback: Callable | None = None) -> bool:
        args = (dbus_id, nid, app_name, replaces_nid, app_icon, summary, body, actions, hints, expire_timeout, icon)
        log("notify%s types=%s", args, tuple(type(x) for x in args))
        if not self.send_notifications:
            log("client %s does not support notifications", self)
            return False
        # "suspended" belongs in the WindowsMixin:
        if getattr(self, "suspended", False):
            log("client %s is suspended, notification not sent", self)
            return False
        if user_callback:
            self.notification_callbacks[nid] = user_callback
        if self.hello_sent:
            # Warning: actions and hints are send last because they were added later (in version 2.3)
            self.send_async("notify_show", dbus_id, nid, app_name, replaces_nid, app_icon,
                            summary, body, expire_timeout, icon or b"", actions, hints)
        return True

    def notify_close(self, nid: int) -> None:
        if not self.send_notifications or self.suspended or not self.hello_sent:
            return
        self.send_more("notify_close", nid)
